import torch

class ExtendedTorchModule(torch.nn.Module):
  def __init__(self,model):
    super().__init__()
    self.model = model

  def train(self,dataloader_train, criterion, optimizer, epoch_count = 10):
    for epoch in range(epoch_count):  # loop over the dataset multiple times

      running_loss = 0.0
      for i, data in enumerate(dataloader_train, 0):
          # get the inputs; data is a list of [inputs, labels]
          inputs, labels = data
          
          # zero the parameter gradients
          optimizer.zero_grad()

          # forward + backward + optimize
          outputs = self(inputs)
          loss = criterion(outputs, labels)
          loss.backward()
          optimizer.step()

          # print statistics
          running_loss += loss.item()
          if i % 5000 == 4999:    # print every 2000 mini-batches
              print(f'[{epoch + 1}, {i + 1:5d}] loss: {running_loss / 5000:.3f}')
              running_loss = 0.0
    print('Finished Training')
    
  def forward(self, x):
    x = x.to(torch.float32)
    # for f in self.module_list:
    #     x = f(x)
    # return x
    return self.model(x)
    
  def compute_outputs(self, dataloader_test):
    output_arr = []
    with torch.no_grad():
      for data in dataloader_test:
          inputs, labels = data
          # calculate outputs by running inputs through the network
          outputs = self(inputs)
          output_arr.append(outputs.numpy())

    # <ipython-input-44-114ac3037693>:54: UserWarning: Creating a tensor from a list of numpy.ndarrays is extremely slow. Please consider converting the list to a single numpy.ndarray with numpy.array() before converting to a tensor. (Triggered internally at ../torch/csrc/utils/tensor_new.cpp:245.)
    t = torch.tensor(output_arr) 
    return torch.squeeze(t) 

  # def compute_accuracy(self, dataloader_test):
  #   correct = 0
  #   total = 0
  #   # since we're not training, we don't need to calculate the gradients for our outputs
  #   with torch.no_grad():
  #       for data in dataloader_test:
  #           inputs, labels = data
  #           # calculate outputs by running inputs through the network
  #           outputs = self(inputs)
  #           # the class with the highest energy is what we choose as prediction
  #           _, predicted = torch.max(outputs.data, 1)
  #           total += labels.size(0)
  #           correct += (predicted == labels).sum().item()

  #   print(f'Accuracy of the network: {100 * correct // total} %')


import glob
from sklearn.model_selection import KFold, StratifiedKFold, train_test_split
from sklearn.preprocessing import LabelBinarizer
from abc import ABC, abstractmethod
import numpy as np
import pandas as pd
import pathlib
from ddi_fw.datasets.idf_helper import IDF

from ddi_fw.utils.zip_helper import ZipHelper
from .feature_vector_generation import SimilarityMatrixGenerator, VectorGenerator
from ddi_fw.ner.ner import CTakesNER
from ddi_fw.utils import create_folder_if_not_exists
from stopwatch import Stopwatch, profile

HERE = pathlib.Path(__file__).resolve().parent


def stack(df_column):
    return np.stack(df_column.values)


class BaseDataset(ABC):
    def __init__(self, chemical_property_columns, embedding_columns, ner_columns,
                 **kwargs):
        self.__similarity_related_columns__ = []
        self.__similarity_related_columns__.extend(chemical_property_columns)
        self.__similarity_related_columns__.extend(ner_columns)

        self.chemical_property_columns = chemical_property_columns
        self.embedding_columns = embedding_columns
        self.ner_columns = ner_columns
        self.threshold_method = kwargs.get('threshold_method', 'idf')
        self.tui_threshold = kwargs.get('tui_threshold', 0)
        self.cui_threshold = kwargs.get('cui_threshold', 0)
        self.entities_threshold = kwargs.get('entities_threshold', 0)

        self.stopwatch = Stopwatch()

        # self.store_similarity_matrices = kwargs.get('store_similarity_matrices', True)
        # self.similarity_matrices_path = kwargs.get('similarity_matrices_path', True)

# önce load veya split çalıştırılmalı
    def produce_inputs(self):
        items = []
        y_train_label, y_test_label = stack(self.y_train), stack(self.y_test)
        for column in self.__similarity_related_columns__:
            train_data, test_data = stack(
                self.X_train[column]), stack(self.X_test[column])
            items.append([f'{column}', np.nan_to_num(train_data),
                          y_train_label, np.nan_to_num(test_data), y_test_label])
        for column in self.embedding_columns:
            train_data, test_data = stack(
                self.X_train[column+'_embedding']), stack(self.X_test[column+'_embedding'])
            items.append([f'{column}_embedding', train_data,
                          y_train_label, test_data, y_test_label])
        return items

##remove this function
    def generate_sim_matrices(self, chemical_properties_df, two_d_dict):

        jaccard_sim_dict = {}
        sim_matrix_gen = SimilarityMatrixGenerator()

        for column in self.__similarity_related_columns__:
            key = '2D_'+column
            jaccard_sim_dict[column] = sim_matrix_gen.create_jaccard_similarity_matrices(
                two_d_dict[key])

        drugbank_ids = chemical_properties_df['id'].to_list()

        similarity_matrices = {}

        for column in self.__similarity_related_columns__:
            sim_matrix = jaccard_sim_dict[column]
            jaccard_sim_feature = {}
            for i in range(len(drugbank_ids)):
                jaccard_sim_feature[drugbank_ids[i]] = sim_matrix[i]
            similarity_matrices[column] = jaccard_sim_feature

        return similarity_matrices

    def generate_sim_matrices_new(self, chemical_properties_df):
        self.stopwatch.reset()
        self.stopwatch.start()
        jaccard_sim_dict = {}
        sim_matrix_gen = SimilarityMatrixGenerator()

        for column in self.__similarity_related_columns__:
            # key = '2D_'+column
            key = column
            jaccard_sim_dict[column] = sim_matrix_gen.create_jaccard_similarity_matrices(
                self.generated_vectors[key])
        self.stopwatch.stop()
        print(f'similarity_matrix_generation_part_1: {self.stopwatch.elapsed}')

        self.stopwatch.reset()
        self.stopwatch.start()
        similarity_matrices = {}
        drugbank_ids = chemical_properties_df['id'].to_list()
        new_columns = {}
        for idx in range(len(drugbank_ids)):
            new_columns[idx] = drugbank_ids[idx]
        for column in self.__similarity_related_columns__:
            new_df = pd.DataFrame.from_dict(jaccard_sim_dict[column])
            new_df = new_df.rename(index=new_columns, columns=new_columns)
            similarity_matrices[column] = new_df
        self.stopwatch.stop()
        print(f'similarity_matrix_generation_part_2: {self.stopwatch.elapsed}')
        return similarity_matrices

    # matris formuna çevirmek için
    def transform_2d(self, chemical_properties_df):
        two_d_dict = {}
        for column in self.__similarity_related_columns__:
            key = '2D_'+column
            new_column = column + '_vectors'
            two_d_dict[key] = np.stack(
                chemical_properties_df[new_column].to_numpy())

        return two_d_dict

    #todo dictionary içinde ndarray dönsün
    def generate_vectors(self, chemical_properties_df):
        self.stopwatch.reset()
        self.stopwatch.start()
        vectorGenerator = VectorGenerator(chemical_properties_df)

        new_columns = [
            c+'_vectors' for c in self.__similarity_related_columns__]
        self.generated_vectors = vectorGenerator.generate_feature_vectors(
            self.__similarity_related_columns__)

        # for column, new_column in zip(self.__similarity_related_columns__, new_columns):
        #     chemical_properties_df.loc[:,
        #                                new_column] = generated_vectors[column]
        # self.generated_vectors = generated_vectors
        self.stopwatch.stop()
        print(f'vector_generation: {self.stopwatch.elapsed}')


##remove this function
    def sim(self,chemical_properties_df):
        self.stopwatch.reset()
        self.stopwatch.start()
        from scipy.spatial.distance import pdist        
        sim_matrix_gen = SimilarityMatrixGenerator()

        drugbank_ids = chemical_properties_df['id'].to_list()
        similarity_matrices = {}
        for column in self.__similarity_related_columns__:
            df = pd.DataFrame(np.stack(chemical_properties_df[f'{column}_vectors'].values), index = drugbank_ids)
        #   similarity_matrices[column] = 1 - pdist(df.to_numpy(), metric='jaccard')
            similarity_matrices[column] = sim_matrix_gen.create_jaccard_similarity_matrices(df.to_numpy())
        self.stopwatch.stop()
        print(f'sim: {self.stopwatch.elapsed}')
        return similarity_matrices

# import pandas as pd
# a = [[0,0,1],[0,0,1],[0,0,0]]
# s = pd.Series(a)
# # print(np.vstack(s.to_numpy()))
# l = np.argmax(np.vstack(s.to_numpy()),axis = 1)
# l
    def split_dataset(self,
                      fold_size=5,
                      shuffle=True,
                      test_size=0.2,
                      save_indexes=False):
        save_path = self.index_path
        self.prep()
        X = self.dataframe.drop('class', axis=1)
        y = self.dataframe['class']
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, shuffle=shuffle, test_size=test_size, stratify=np.argmax(np.vstack(y.to_numpy()),axis = 1))
        # k_fold = KFold(n_splits=fold_size, shuffle=shuffle, random_state=1)
        # folds = k_fold.split(X_train)

        k_fold = StratifiedKFold(n_splits=fold_size, shuffle=shuffle, random_state=1)
        folds = k_fold.split(X_train, np.argmax(np.vstack(y_train.to_numpy()),axis = 1))
        train_idx_arr = []
        val_idx_arr = []
        for i, (train_index, val_index) in enumerate(folds):
            train_idx_arr.append(train_index)
            val_idx_arr.append(val_index)

        if save_indexes:
            # train_pairs = [row['id1'].join(',').row['id2'] for index, row in X_train.iterrows()]
            self.__save_indexes__(
                save_path, 'train_indexes.txt', X_train['index'].values)
            self.__save_indexes__(
                save_path, 'test_indexes.txt',  X_test['index'].values)
            # self.__save_indexes__(
            #     save_path, 'train_indexes.txt', X_train.index.values)
            # self.__save_indexes__(
            #     save_path, 'test_indexes.txt',  X_test.index.values)

            for i, (train_idx, val_idx) in enumerate(zip(train_idx_arr, val_idx_arr)):
                self.__save_indexes__(
                    save_path, f'train_fold_{i}.txt', train_idx)
                self.__save_indexes__(
                    save_path, f'validation_fold_{i}.txt', val_idx)

        self.X_train = X_train
        self.X_test = X_test
        self.y_train = y_train
        self.y_test = y_test
        self.train_indexes = X_train.index
        self.test_indexes = X_test.index
        self.train_idx_arr = train_idx_arr
        self.val_idx_arr = val_idx_arr
        return X_train, X_test, y_train, y_test, X_train.index, X_test.index, train_idx_arr, val_idx_arr

    def __get_indexes__(self, path):
        train_index_path = path+'/train_indexes.txt'
        test_index_path = path+'/test_indexes.txt'
        train_fold_files = f'{path}/train_fold_*.txt'
        val_fold_files = f'{path}/validation_fold_*.txt'
        train_idx_arr = []
        val_idx_arr = []
        with open(train_index_path, 'r', encoding="utf8") as f:
            train_idx_all = [int(r) for r in f.readlines()]
        with open(test_index_path, 'r', encoding="utf8") as f:
            test_idx_all = [int(r) for r in f.readlines()]

        for filepath in glob.glob(train_fold_files):
            with open(filepath, 'r', encoding="utf8") as f:
                train_idx = [int(r) for r in f.readlines()]
                train_idx_arr.append(train_idx)
        for filepath in glob.glob(val_fold_files):
            with open(filepath, 'r', encoding="utf8") as f:
                val_idx = [int(r) for r in f.readlines()]
                val_idx_arr.append(val_idx)
        return train_idx_all, test_idx_all, train_idx_arr, val_idx_arr

    def __save_indexes__(self, path, filename, indexes):
        create_folder_if_not_exists(path)
        file_path = path + '/'+filename
        str_indexes = [str(index) for index in indexes]
        with open(file_path, 'w') as f:
            f.write('\n'.join(str_indexes))

    # @abstractmethod
    # def prep(self):
    #     pass

    # @abstractmethod
    # def load(self):
    #     pass

# her bir metin tipi için embedding oluşturursan burayı düzenle
    def prep(self):
        if self.embedding_columns:
            zip_helper = ZipHelper()
            zip_helper.extract(str(HERE.joinpath('zips/embeddings')),
                               str(HERE.joinpath('zips/embeddings')))

        embedding_dict = dict()
        for embedding_column in self.embedding_columns:
            embedding_file = HERE.joinpath(
                f'zips/embeddings/{embedding_column}_embeddings.pkl')
            embedding_values = pd.read_pickle(embedding_file)
            d = embedding_values.apply(
                lambda x: {x.id: x[f'{embedding_column}_embedding']}, axis=1)
            x = {k: v for l in d.values.tolist() for k, v in l.items()}
            embedding_dict[embedding_column] = x

        self.ner_df = CTakesNER().load()
        drug_names = self.drugs_df['name'].to_list()
        drug_ids = self.drugs_df['id'].to_list()

        # self.ddis_df = self.ddis_df[(self.ddis_df['name1'].isin(drug_names)) & (
        #     self.ddis_df['name2'].isin(drug_names))]

        filtered_df = self.drugs_df
        filtered_ner_df = self.ner_df[self.ner_df['drugbank_id'].isin(
            drug_ids)]
        filtered_ner_df = self.ner_df.copy()

        combined_df = filtered_df.copy()
        # TODO: eğer kullanılan veri setinde tui, cui veya entity bilgileri yoksa o veri setine bu sütunları eklemek için aşağısı gerekli

        idf_calc = IDF(filtered_ner_df, [f for f in filtered_ner_df.keys()])
        idf_calc.calculate()
        idf_scores_df = idf_calc.to_dataframe()

        for key in filtered_ner_df.keys():
            threshold = 0
            if key.startswith('tui'):
                threshold = self.tui_threshold
            if key.startswith('cui'):
                threshold = self.cui_threshold
            if key.startswith('entities'):
                threshold = self.entities_threshold
            combined_df[key] = filtered_ner_df[key]
            valid_codes = idf_scores_df[idf_scores_df[key] > threshold].index

            # print(f'{key}: valid code size = {len(valid_codes)}')
            combined_df[key] = combined_df[key].apply(lambda items:
                                                      [item for item in items if item in valid_codes])

        # Yukarıdaki koda evrildi
        # combined_df['tui_description'] = filtered_ner_df['description_tuis']
        # combined_df['cui_description'] = filtered_ner_df['description_cuis']
        # combined_df['entities_description'] = filtered_ner_df['description_entities']

        # tui_idf = IDF(combined_df['tui_description'], self.tui_threshold)
        # cui_idf = IDF(combined_df['cui_description'], self.cui_threshold)
        # entities_idf = IDF(
        #     combined_df['entities_description'], self.entities_threshold)

        # tui_idf.calculate()
        # cui_idf.calculate()
        # entities_idf.calculate()

        # valid_tui_codes = tui_idf.find_items_over_threshold()
        # valid_cui_codes = cui_idf.find_items_over_threshold()
        # valid_entities_codes = entities_idf.find_items_over_threshold()

        # combined_df['tui_description'] = combined_df['tui_description'].apply(lambda items:
        #                                                                       [item for item in items if item in valid_tui_codes])
        # combined_df['cui_description'] = combined_df['cui_description'].apply(lambda items:
        #                                                                       [item for item in items if item in valid_cui_codes])
        # combined_df['entities_description'] = combined_df['entities_description'].apply(lambda items:
        #                                                                                 [item for item in items if item in valid_entities_codes])

        moved_columns = ['id']
        moved_columns.extend(self.__similarity_related_columns__)
        chemical_properties_df = combined_df[moved_columns]

        chemical_properties_df = chemical_properties_df.fillna("").apply(list)

        # generate vectors dictionary içinde ndarray dönecek
        self.generate_vectors(chemical_properties_df)

        # two_d_dict = self.transform_2d(chemical_properties_df)

        similarity_matrices = self.generate_sim_matrices_new(
            chemical_properties_df)

        # similarity_matrices = self.sim(chemical_properties_df)

        event_categories = self.ddis_df['event_category']
        labels = event_categories.tolist()
        lb = LabelBinarizer()
        lb.fit(labels)
        classes = lb.transform(labels)

        # def similarity_lambda_fnc(row, value):
        #     if row['id1'] in value and row['id2'] in value:
        #         return value[row['id1']][row['id2']]
            
        def similarity_lambda_fnc(row, value):
            if row['id1'] in value:
                return value[row['id1']]

        def lambda_fnc(row, value):
            if row['id1'] in value and row['id2'] in value:
                return np.float16(np.hstack(
                    (value[row['id1']], value[row['id2']])))
                # return np.hstack(
                #     (value[row['id1']], value[row['id2']]), dtype=np.float16)

        def x_fnc(row, embedding_values, embedding_column):
            # first = embedding_values[embedding_values.id == row['id1']]
            # second = embedding_values[embedding_values.id == row['id2']]
            # v1 = first.iloc[0][embedding_column+'_embedding']
            # v2 = second.iloc[0][embedding_column+'_embedding']
            v1 = embedding_dict[embedding_column][row['id1']]
            v2 = embedding_dict[embedding_column][row['id2']]
            # v1 = embedding_dict[row['id1']][embedding_column+'_embedding']
            # v2 = embedding_dict[row['id2']][embedding_column+'_embedding']
            return np.float16(np.hstack(
                (v1, v2)))

        for key, value in similarity_matrices.items():

            print(f'sim matrix: {key}')
            self.ddis_df[key] = self.ddis_df.apply(
                lambda_fnc, args=(value,), axis=1)
            print(self.ddis_df[key].head())

        for embedding_column in self.embedding_columns:
            print(f"concat {embedding_column} embeddings")
            # column_embeddings_dict = embedding_values[embedding_column]
            self.ddis_df[embedding_column+'_embedding'] = self.ddis_df.apply(
                x_fnc, args=(embedding_values, embedding_column), axis=1)

        self.dataframe = self.ddis_df.copy()
        self.dataframe['class'] = list(classes)
        print(self.dataframe.shape)

    def load(self):
        if self.index_path == None:
            raise Exception(
                "There is no index path, please call split function")

        # prep - split - load
        train_idx_all, test_idx_all, train_idx_arr, val_idx_arr = self.__get_indexes__(
            self.index_path)

        self.prep()
        train = self.dataframe[self.dataframe['index'].isin(train_idx_all)]
        test = self.dataframe[self.dataframe['index'].isin(test_idx_all)]

        self.X_train = train.drop('class', axis=1)
        self.y_train = train['class']
        self.X_test = test.drop('class', axis=1)
        self.y_test = test['class']

        self.train_indexes = self.X_train.index
        self.test_indexes = self.X_test.index
        self.train_idx_arr = train_idx_arr
        self.val_idx_arr = val_idx_arr

        return self.X_train, self.X_test, self.y_train, self.y_test, self.X_train.index, self.X_test.index, train_idx_arr, val_idx_arr



import sqlite3
import pandas as pd
import os
import json
import glob
from tqdm import tqdm

import csv

from rdkit import Chem
from rdkit.Chem import AllChem
import numpy as np
from ddi_fw.drugbank.event_extractor import EventExtractor
from ddi_fw.ner.ner import CTakesNER

from ddi_fw.utils import ZipHelper
# from event_extractor import EventExtractor


def multiline_to_singleline(multiline):
    if multiline is None:
        return ""
    return " ".join(line.strip() for line in multiline.splitlines())

# targets -> target -> polypeptide
# enzymes -> enzyme -> polypeptide
# pathways from KEGG, KEGG ID is obtained from DrugBank
# https://www.genome.jp/dbget-bin/www_bget?drug:D03136
# https://www.kegg.jp/entry/D03136


class DrugBankProcessor():

    def mask_interaction(self, drug_1, drug_2, interaction):
        return interaction.replace(
            drug_1, "DRUG").replace(drug_2, "DRUG")

    def extract_zip_files(self, input_path='zips', output_path='drugs', override=False):
        if override:
            zip_helper = ZipHelper()
            zip_helper.extract(input_path=input_path, output_path=output_path)

    def get_external_identifiers(self,input_path='drugs'):
        external_identifier_list = []
        all_json_files = input_path+'/*.json*'

        for filepath in tqdm(glob.glob(all_json_files)):
            with open(filepath, 'r', encoding="utf8") as f:

                data = json.load(f)
                drug_1 = data['name']
                drug_1_id = [d['value']
                                     for d in data['drugbank-id'] if d['primary'] == True][0]
                external_identifiers = data['external-identifiers'] if "external-identifiers" in data else None
                external_identifiers_dict = {}
                external_identifiers_dict['name'] = drug_1
                external_identifiers_dict['drugbank_id'] = drug_1_id
                if external_identifiers is not None:
                    for p in external_identifiers['external-identifier']:
                        external_identifiers_dict[p['resource'].lower().replace(" ","_")] = p['identifier']
                    # external_identifiers_dict = dict(
                    #     [(p['resource'].lower().replace(" ","_"), p['identifier']) for p in external_identifiers['external-identifier']])
                    # external_identifiers_dict['name'] = drug_1
                    # external_identifiers_dict['drugbank_id'] = drug_1_id
                external_identifier_list.append(external_identifiers_dict)
        return external_identifier_list


    def process(self, input_path='drugs', output_path='output', save_as_sql=True, db_path =  r"./drugbank.db", zip_outputs=True):
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        ner_df =CTakesNER().load()
        drugs_pickle_path = output_path+'/drugs.pkl'
        drugs_csv_path = output_path+'/drugs.gzip'
        ddi_pickle_path = output_path + '/ddi.pkl'
        ddi_csv_path = output_path + '/ddi.gzip'

        if not os.path.exists(drugs_pickle_path) or not os.path.exists(ddi_pickle_path):
            drug_rows = []
            all_ddis = []
            external_identifier_list = []
            all_json_files = input_path+'/*.json*'

            for filepath in tqdm(glob.glob(all_json_files)):
                with open(filepath, 'r', encoding="utf8") as f:

                    data = json.load(f)

                    # if data['drug-interactions'] is None:
                    if False:
                        continue
                    else:
                        drug_1 = data['name']
                        drug_1_id = [d['value']
                                     for d in data['drugbank-id'] if d['primary'] == True][0]
                        description = multiline_to_singleline(
                            data['description'])
                        if data['drug-interactions'] is not None:
                            drug_interactions = [
                                interaction for interaction in data['drug-interactions']['drug-interaction']]
                            ddis = [(drug_1, interaction['name'], interaction['description'])
                                    for interaction in data['drug-interactions']['drug-interaction']]

                            ddi_dict = [{
                                'drug_1_id': drug_1_id,
                                'drug_1': drug_1,
                                'drug_2_id': interaction['drugbank-id']['value'],
                                'drug_2': interaction['name'],
                                'interaction': interaction['description'],
                                'masked_interaction': self.mask_interaction(drug_1, interaction['name'], interaction['description'])}
                                for interaction in data['drug-interactions']['drug-interaction']]
                            all_ddis.extend(ddi_dict)

                        synthesis_reference = data['synthesis-reference']
                        indication = multiline_to_singleline(
                            data['indication'])
                        pharmacodynamics = multiline_to_singleline(
                            data['pharmacodynamics'])
                        mechanism_of_action = multiline_to_singleline(
                            data['mechanism-of-action'])
                        toxicity = multiline_to_singleline(data['toxicity'])
                        metabolism = multiline_to_singleline(
                            data['metabolism'])
                        absorption = multiline_to_singleline(
                            data['absorption'])
                        half_life = multiline_to_singleline(data['half-life'])
                        protein_binding = multiline_to_singleline(
                            data['protein-binding'])
                        route_of_elimination = multiline_to_singleline(
                            data['route-of-elimination'])
                        volume_of_distribution = multiline_to_singleline(
                            data['volume-of-distribution'])
                        clearance = multiline_to_singleline(data['clearance'])

                        food_interactions = data['food-interactions']
                        sequences = data['sequences'] if "sequences" in data else None

                        external_identifiers = data['external-identifiers'] if "external-identifiers" in data else None
                        experimental_properties = data['experimental-properties'] if "experimental-properties" in data else None
                        calculated_properties = data['calculated-properties'] if "calculated-properties" in data else None

                        enzymes_polypeptides = None
                        targets_polypeptides = None
                        pathways = None

                        # targets = data['targets'] if "targets" in data else None
                        if data['targets'] is not None:
                            # targets_polypeptides = [p['id'] for d in data['targets']['target'] for p in d['polypeptide'] if 'polypeptide' in d ]
                            targets_polypeptides = [
                                p['id'] for d in data['targets']['target'] if 'polypeptide' in d for p in d['polypeptide']]

                        if data['enzymes'] is not None:
                            # enzymes_polypeptides = [p['id'] for d in data['enzymes']['enzyme'] for p in d['polypeptide'] if 'polypeptide' in d]
                            enzymes_polypeptides = [
                                p['id'] for d in data['enzymes']['enzyme'] if 'polypeptide' in d for p in d['polypeptide']]

                        if data['pathways'] is not None:
                            pathways = [
                                d['smpdb-id'] for d in data['pathways']['pathway']]


                        if external_identifiers is not None:
                            external_identifiers_dict = dict(
                                [(p['resource'], p['identifier']) for p in external_identifiers['external-identifier']])
                            external_identifiers_dict['drugbank_id'] = drug_1_id
                            external_identifier_list.append(external_identifiers_dict)
# add note column
                        smiles = None
                        morgan_hashed = None
                        if calculated_properties is not None:
                            calculated_properties_dict = dict(
                                [(p['kind'], p['value']) for p in calculated_properties['property']])
                            smiles = calculated_properties_dict['SMILES'] if 'SMILES' in calculated_properties_dict else None
                            if smiles is not None:
                                try:
                                    mol = Chem.MolFromSmiles(smiles)
                                    morgan_hashed = AllChem.GetMorganFingerprintAsBitVect(
                                        mol, 2, nBits=881).ToList()
                                except:
                                    print("An exception occurred")
                        if morgan_hashed is None:
                            morgan_hashed = np.zeros(881).tolist()

                        # TODO cui, tui, entities other types of texts, test it
                        tuis_description = ner_df[ner_df['drugbank_id'] == drug_1_id]['tui_description'].values
                        if len(tuis_description) > 0:
                            tuis_description = tuis_description[0]
                        else:
                            tuis_description = None

                        cuis_description = ner_df[ner_df['drugbank_id'] == drug_1_id]['cui_description'].values
                        if len(cuis_description) > 0:
                            cuis_description = cuis_description[0]
                        else:
                            cuis_description = None

                        entities_description = ner_df[ner_df['drugbank_id'] == drug_1_id]['entities_description'].values
                        if len(entities_description) > 0:
                            entities_description = entities_description[0]
                        else:
                            entities_description = None

                        # k = [p[k]  for p in calculated_properties['property'] for k in p.keys() if k =='SMILES']
                        # external_identifiers['external-identifier']
                        # experimental_properties['property']

# list to single line reminder
                        row = {'drugbank_id': drug_1_id,
                               'name': drug_1,
                               'description': description,
                               'synthesis_reference': synthesis_reference,
                               'indication': indication,
                               'pharmacodynamics': pharmacodynamics,
                               'mechanism_of_action': mechanism_of_action,
                               'toxicity': toxicity,
                               'metabolism': metabolism,
                               'absorption': absorption,
                               'half_life': half_life,
                               'protein_binding': protein_binding,
                               'route_of_elimination': route_of_elimination,
                               'volume_of_distribution': volume_of_distribution,
                               'clearance': clearance,
                               'smiles': smiles,
                               'smiles_morgan_fingerprint': ','.join(map(str, morgan_hashed)),
                               'enzymes_polypeptides': '|'.join(enzymes_polypeptides) if enzymes_polypeptides is not None else None,
                               'targets_polypeptides': '|'.join(targets_polypeptides) if targets_polypeptides is not None else None,
                               'pathways': '|'.join(pathways) if pathways is not None else None,
                               'tuis_description':'|'.join(tuis_description) if tuis_description is not None else None,
                               'cuis_description':'|'.join(cuis_description) if cuis_description is not None else None,
                               'entities_description':'|'.join(entities_description) if entities_description is not None else None
                            #    'external_identifiers': external_identifiers_dict
                               }
                        drug_rows.append(row)

                    # if len(drug_rows) == 10:
                    #     break
            # print(smiles_count)
            print(f"Size of drugs {len(drug_rows)}")
            print(f"Size of DDIs {len(all_ddis)}")
            np.set_printoptions(threshold=np.inf)

            # drug_names = [row['name'] for row in drug_rows]
            drug_names = ['DRUG']
            event_extractor = EventExtractor(drug_names)

            replace_dict = {'MYO-029': 'Stamulumab'}
            for ddi in tqdm(all_ddis):
                for key, value in replace_dict.items():
                    ddi['masked_interaction'] = ddi['masked_interaction'].replace(
                        key, value)
            #     interaction = ddi['interaction']
            #     mechanism, action, drugA, drugB = event_extractor.extract(interaction)
            #     ddi['mechanism'] = mechanism
            #     ddi['action'] = action

            self.drugs_df = pd.DataFrame(drug_rows)
            # self.drugs_df.to_pickle(drugs_pickle_path)
            # self.drugs_df.to_csv(
            #     drugs_csv_path,  index=False, compression='gzip')

            # print('mechanism_action calculation')
            self.ddis_df = pd.DataFrame(all_ddis)

            count = [0]

            def fnc2(interaction, count):
                count[0] = count[0] + 1
                if count[0] % 1000 == 0:
                    print(f'{count[0]}/{len(all_ddis)}')
                mechanism, action, drugA, drugB = event_extractor.extract(
                    interaction)
                return mechanism+'__' + action

            # self.ddis_df['mechanism_action'] = self.ddis_df['interaction'].apply(lambda x: fnc2(x))
            # tqdm.pandas()
            self.ddis_df['mechanism_action'] = self.ddis_df['masked_interaction'].apply(
                fnc2, args=(count,))

            # self.ddis_df.to_csv(ddi_csv_path,  index=False, compression='gzip')
            # self.ddis_df.to_pickle(ddi_pickle_path)
            zip_helper = ZipHelper()

            if save_as_sql:
                conn = sqlite3.connect(db_path)
                # self.drugs_df.to_sql('_Drugs', conn, if_exists='replace', index=True)
                # self.ddis_df.to_sql('_Interactions', conn, if_exists='replace', index=True)
                ext_id_df= pd.DataFrame.from_records(external_identifier_list)
                ext_id_df.to_sql('_ExternalIdentifiers', conn, if_exists='replace', index=True)

                zip_helper.zip_single_file(
                    file_path=db_path, output_path=output_path+'/zips', name='db')
                conn.close()

            if zip_outputs:
                zip_helper.zip_single_file(
                    file_path=drugs_pickle_path, output_path=output_path+'/zips', name='drugs-pickle')
                zip_helper.zip_single_file(
                    file_path=ddi_pickle_path, output_path=output_path+'/zips', name='ddi-pickle')

        else:
            print('Output path has processed data, load function is called')
            self.load(output_path)

    def load(self, path):
        drugs_pickle_path = path+'/drugs.pkl'
        ddi_pickle_path = path+'/ddi.pkl'
        if os.path.exists(drugs_pickle_path) and os.path.exists(ddi_pickle_path):
            self.drugs_df = pd.read_pickle(drugs_pickle_path)
            self.ddis_df = pd.read_pickle(ddi_pickle_path)
        else:
            print('One of given paths could not found')

    def load_from_csv(self, path):
        drugs_csv_path = path+'/drugs.gzip'
        ddi_csv_path = path+'/ddi.gzip'
        if os.path.exists(drugs_csv_path) and os.path.exists(ddi_csv_path):
            self.drugs_df = pd.read_csv(drugs_csv_path, compression='gzip')
            self.ddis_df = pd.read_csv(ddi_csv_path, compression='gzip')
        else:
            print('One of given paths could not found')

    def load2(self, path):
        drugs_pickle_path = path+'/drugs.pkl'
        ddi_csv_path = path+'/ddi.gzip'
        if os.path.exists(drugs_pickle_path) and os.path.exists(ddi_csv_path):
            self.drugs_df = pd.read_pickle(drugs_pickle_path)
            self.ddis_df = pd.read_csv(ddi_csv_path, compression='gzip')
        else:
            print('One of given paths could not found')

    def drugs_as_dataframe(self):
        return self.drugs_df

    def filtered_drugs_as_dataframe(self, drug_ids):
        return self.drugs_df[self.drugs_df['drugbank_id'].isin(drug_ids)]

    def ddis_as_dataframe(self):
        return self.ddis_df

    def filtered_ddis(self, drugs):
        ddis_df = self.ddis_df.copy()
        return ddis_df[(ddis_df['drug_1'] in drugs) & (
            ddis_df['drug_2'] in drugs)]



# # https://github.com/kashif/tf-keras-tutorial/blob/tf2/3-imdb.ipynb
# # TensorFlow and tf.keras
# import tensorflow as tf

# # Helper libraries
# import numpy as np
# import matplotlib.pyplot as plt
# from tensorflow_helper import CustomCallback

# print(tf.__version__)


# imdb = tf.keras.datasets.imdb

# (train_data, train_labels), (test_data, test_labels) = tf.keras.datasets.imdb.load_data(num_words=10000)


# class_names = ['T-shirt/top', 'Trouser', 'Pullover', 'Dress', 'Coat',
#                'Sandal', 'Shirt', 'Sneaker', 'Bag', 'Ankle boot']

  
# # Create a model
# model = tf.keras.Sequential()
# custom_callback = CustomCallback()

#  # input shape here is the length of our movie review vector
# model.add(tf.keras.layers.Dense(16, activation=tf.nn.relu, input_shape=(10000,)))
# model.add(tf.keras.layers.Dense(16, activation=tf.nn.relu))
# model.add(tf.keras.layers.Dense(1, activation=tf.nn.sigmoid))

# optimizer = tf.keras.optimizers.RMSprop(learning_rate=0.001)

# model.compile(loss='binary_crossentropy',
#               optimizer=optimizer,
#               metrics=['binary_accuracy'])

# model.summary()
 
# VAL_SIZE = 10000
# x  = np.array(train_data[:VAL_SIZE].tolist())

# val_data = np.asarray(train_data[:VAL_SIZE])
# partial_train_data = np.asarray(train_data[VAL_SIZE:])


# val_labels = train_labels[:VAL_SIZE]
# partial_train_labels = train_labels[VAL_SIZE:]

# BATCH_SIZE = 512
# SHUFFLE_SIZE = 1000

# # training_set = tf.data.Dataset.from_tensor_slices((partial_train_data, partial_train_labels))
# # training_set = training_set.shuffle(SHUFFLE_SIZE).batch(BATCH_SIZE)

# model.fit(partial_train_data , partial_train_labels , batch_size=128, epochs=20, validation_data=(val_data , val_labels ),
#                           callbacks=[custom_callback])

# loss, accuracy = model.evaluate(test_data, test_labels,callbacks=[custom_callback])
# print('Test accuracy: %.2f' % (accuracy))


import numpy as np
from sklearn import metrics
from sklearn.metrics import accuracy_score, precision_recall_curve
from sklearn.metrics import f1_score
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import roc_auc_score
from sklearn.metrics import auc
from sklearn.metrics import classification_report
from sklearn.preprocessing import OneHotEncoder

def __format__(d,floating_number_precision = 4):
    if type(d) is dict:
        d = {k: __round__(v,floating_number_precision) for k, v in d.items()}
    else:
        d = round(d,floating_number_precision)
    return d

def __round__(v,floating_number_precision = 4):
    if type(v) is list or type(v) is set:
       return [round(item,floating_number_precision) for item in v]
    else:
        return round(v,floating_number_precision)


class Metrics():
    def __init__(self, label):
        self.label = label

    def classification_report(self,classification_report):
        self.classification_report = classification_report

    def accuracy(self, accuracy):
        self.accuracy = accuracy

    def precision(self, precision):
        self.precision = precision

    def recall(self, recall):
        self.recall = recall

    def f1_score(self, f1_score):
        self.f1_score = f1_score

    def roc_auc(self, roc_auc):
        self.roc_auc = roc_auc

    def roc_aupr(self, roc_aupr):
        self.roc_aupr = roc_aupr

    def format_float(self, floating_number_precision = 4):
        self.accuracy = round(self.accuracy,floating_number_precision)
        self.precision = __format__( self.precision ,floating_number_precision)
        self.recall = __format__( self.recall ,floating_number_precision)
        self.f1_score = __format__( self.f1_score ,floating_number_precision)
        self.roc_auc = __format__( self.roc_auc ,floating_number_precision)
        self.roc_aupr = __format__( self.roc_aupr ,floating_number_precision)


# taken from https://github.com/YifanDengWHU/DDIMDL/blob/master/DDIMDL.py#L214
def roc_aupr_score(y_true, y_score, average="macro"):
    def _binary_roc_aupr_score(y_true, y_score):
        precision, recall, pr_thresholds = precision_recall_curve(
            y_true, y_score)
        # precision, recall, pr_thresholds = metrics.roc_curve(y, pred, pos_label=2)
        return auc(precision, recall)

    def _average_binary_score(binary_metric, y_true, y_score, average):  # y_true= y_one_hot
        if average == "binary":
            return binary_metric(y_true, y_score)
        if average == "micro":
            y_true = y_true.ravel()
            y_score = y_score.ravel()
        if y_true.ndim == 1:
            y_true = y_true.reshape((-1, 1))
        if y_score.ndim == 1:
            y_score = y_score.reshape((-1, 1))
        n_classes = y_score.shape[1]
        score = np.zeros((n_classes,))
        for c in range(n_classes):
            y_true_c = y_true.take([c], axis=1).ravel()
            y_score_c = y_score.take([c], axis=1).ravel()
            score[c] = binary_metric(y_true_c, y_score_c)
        return np.average(score)

    return _average_binary_score(_binary_roc_aupr_score, y_true, y_score, average)

# actual and pred are one-hot encoded


def evaluate(actual, pred, info = '' ,print=False):
    
    y_pred = np.argmax(pred, axis=1)
    y_true = np.argmax(actual, axis=1)
    c_report = classification_report(y_true, y_pred, output_dict = True)
    

    metrics = Metrics(info)

    precision = dict()
    recall = dict()
    f_score = dict()
    roc_aupr = dict()
    roc_auc = dict()

    # Compute Precision-Recall and ROC-AUPR for each class
    for i in range(actual.shape[1]):
        precision[i], recall[i], _ = precision_recall_curve(
            actual[:, i].ravel(), pred[:, i].ravel())
        roc_aupr[i] = auc(recall[i], precision[i])
        precision[i] = precision[i].tolist()
        recall[i] = recall[i].tolist()
        classes = [1 if i == np.argmax(y) else 0 for y in y_true]
        # roc_auc[i] = roc_auc_score(classes, pred[:,i])

    roc_auc["weighted"] = roc_auc_score(
        actual, pred, multi_class='ovr', average='weighted')
    roc_auc["macro"] = roc_auc_score(
        actual, pred, multi_class='ovr', average='macro')
    roc_auc["micro"] = roc_auc_score(
        actual, pred, multi_class='ovr', average='micro')

    # Compute micro-average Precision-Recall curve and ROC-AUPR
    precision["micro_event"], recall["micro_event"], _ = precision_recall_curve(
        actual.ravel(), pred.ravel())
    roc_aupr["micro"] = auc(recall["micro_event"], precision["micro_event"])
    precision["micro_event"] = precision["micro_event"].tolist()
    recall["micro_event"] = recall["micro_event"].tolist()
    # weighted_roc_auc_score = roc_auc_score(actual, pred, multi_class='ovr', average='weighted')
    # macro_roc_auc_score = roc_auc_score(actual, pred, multi_class='ovr', average='macro')
    # micro_roc_auc_score = roc_auc_score(actual, pred, multi_class='ovr', average='micro')

    # macro_aupr_score = roc_aupr_score(actual, pred, average='macro')
    # micro_aupr_score = roc_aupr_score(actual, pred, average='micro')

    acc = accuracy_score(y_true, y_pred)

    precision['weighted'] = precision_score(y_true, y_pred, average='weighted')
    precision['macro'] = precision_score(y_true, y_pred, average='macro')
    precision['micro'] = precision_score(y_true, y_pred, average='micro')

    recall['weighted'] = recall_score(y_true, y_pred, average='weighted')
    recall['macro'] = recall_score(y_true, y_pred, average='macro')
    recall['micro'] = recall_score(y_true, y_pred, average='micro')

    f_score['weighted'] = f1_score(y_true, y_pred, average='weighted')
    f_score['macro'] = f1_score(y_true, y_pred, average='macro')
    f_score['micro'] = f1_score(y_true, y_pred, average='micro')

    # acc = accuracy_score(y_true, y_pred)

    # weighted_precision = precision_score(y_true, y_pred, average='weighted')
    # macro_precision = precision_score(y_true, y_pred, average='macro')
    # micro_precision = precision_score(y_true, y_pred, average='micro')

    # weighted_recall_score = recall_score(y_true, y_pred, average='weighted')
    # macro_recall_score = recall_score(y_true, y_pred, average='macro')
    # micro_recall_score = recall_score(y_true, y_pred, average='micro')

    # weighted_f1_score = f1_score(y_true, y_pred, average='weighted')
    # macro_f1_score = f1_score(y_true, y_pred, average='macro')
    # micro_f1_score = f1_score(y_true, y_pred, average='micro')

    if print:
        print(
            f'''Accuracy: {acc}
            , Precision:{precision['weighted']}
            , Recall: {recall['weighted']}
            , F1-score: {f_score['weighted']}
            ''')

    logs = {'accuracy': acc,
            'weighted_precision': precision['weighted'],
            'macro_precision': precision['macro'],
            'micro_precision': precision['micro'],
            'weighted_recall_score': recall['weighted'],
            'macro_recall_score': recall['macro'],
            'micro_recall_score': recall['micro'],
            'weighted_f1_score': f_score['weighted'],
            'macro_f1_score': f_score['macro'],
            'micro_f1_score': f_score['micro'],
            # 'weighted_roc_auc_score': weighted_roc_auc_score,
            # 'macro_roc_auc_score': macro_roc_auc_score,
            # 'micro_roc_auc_score': micro_roc_auc_score,
            # 'macro_aupr_score': macro_aupr_score,
            # 'micro_aupr_score': micro_aupr_score
            "micro_roc_aupr": roc_aupr['micro'],
            # "micro_precision_from_precision_recall_curve":precision["micro"],
            # "micro_recall_from_precision_recall_curve":recall["micro"],
            "weighted_roc_auc": roc_auc['weighted'],
            "macro_roc_auc": roc_auc['macro'],
            "micro_roc_auc": roc_auc['micro']
            }
    metrics.accuracy(acc)
    metrics.precision(precision)
    metrics.recall(recall)
    metrics.f1_score(f_score)
    metrics.roc_auc(roc_auc)
    metrics.roc_aupr(roc_aupr)
    metrics.classification_report(c_report)
    return logs, metrics


# # Sample integer array
# integer_array = np.array([0, 1, 2, 1, 0])

# # Reshape the integer array to a column vector
# integer_array = integer_array.reshape(-1, 1)

# # Create OneHotEncoder object
# encoder = OneHotEncoder(sparse_output=False)

# # Fit and transform the integer array to one-hot encoded array
# y_true = encoder.fit_transform(integer_array)
# # y_true = np.array([[1, 0, 0],
# #                    [0, 1, 0],
# #                    [0, 0, 1],
# #                    [1, 0, 0],
# #                    [0, 0, 1]],
# #                    )
# y_score = np.array([[0.6, 0.2, 0.2],
#                     [0.2, 0.5, 0.3],
#                     [0.1, 0.2, 0.7],
#                     [0.1, 0.8, 0.1],
#                     [0.1, 0.6, 0.3]])

# y = np.array([-1, -1, 1, 1])
# pred = np.array([0.1, 0.4, 0.35, 0.8])
# evaluate(y_true,y_score)
# fpr, tpr, thresholds = metrics.roc_curve(y, pred)
# print(metrics.auc(fpr, tpr))
# print(roc_aupr_score(y,pred))



from matplotlib import pyplot as plt
import tensorflow as tf
from tensorflow import keras
from keras.models import Model, Sequential
from keras.layers import Dense, Dropout, Input, Activation, BatchNormalization
from keras.callbacks import EarlyStopping
from sklearn.model_selection import train_test_split, KFold, StratifiedKFold
import numpy as np

import mlflow
from mlflow.utils.autologging_utils import batch_metrics_logger
import time

from mlflow.models import infer_signature
from ddi_fw.experiments.evaluation_helper import evaluate

import tf2onnx
import onnx

import itertools
import ddi_fw.utils

# https://github.com/YifanDengWHU/DDIMDL/blob/master/newTask.py
# from numpy.random import seed
# seed(1)
# from tensorflow import set_random_seed
# set_random_seed(2)
tf.random.set_seed(1)
np.random.seed(2)
np.set_printoptions(precision=4)


class TFMultiModal:
    # todo model related parameters to config
    def __init__(self, model_func, batch_size=128, epochs=100):
        self.model_func = model_func
        self.batch_size = batch_size
        self.epochs = epochs

    def set_data(self, items, train_idx_arr, val_idx_arr, y_test_label):
        self.items = items
        self.train_idx_arr = train_idx_arr
        self.val_idx_arr = val_idx_arr
        self.y_test_label = y_test_label

    def predict(self, combinations: list = [], generate_combinations=False):
        self.prefix = utils.utc_time_as_string()
        self.date = utils.utc_time_as_string_simple_format
        sum = np.zeros(
            (self.y_test_label.shape[0], self.y_test_label.shape[1]))
        single_results = dict()

        if generate_combinations:
            l = [item[0] for item in self.items]
            combinations = []
            for i in range(2, len(l) + 1):
                combinations.extend(list(itertools.combinations(l, i)))  # all

        with mlflow.start_run(run_name=self.prefix, description="***") as run:
            self.level_0_run_id = run.info.run_id
            for item in self.items:
                print(item[0])
                single_modal = TFSingleModal(
                    self.date, item[0], self.model_func, self.batch_size, self.epochs)
                single_modal.set_data(
                    self.train_idx_arr, self.val_idx_arr, item[1], item[2], item[3], item[4])
                r = single_modal.predict()
                single_results[item[0]] = r
                sum = sum + r

            if combinations:
                self.evaluate_combinations(single_results, combinations)
        # TODO: sum'a gerek yok
        return sum, single_results

    def evaluate_combinations(self, single_results, combinations):
        for combination in combinations:
            combination_descriptor = '-'.join(combination)
            with mlflow.start_run(run_name=combination_descriptor, description="***", nested=True) as combination_run:
                prediction = np.zeros(
                    (self.y_test_label.shape[0], self.y_test_label.shape[1]))
                for item in combination:
                    prediction = prediction + single_results[item]
                logs, metrics = evaluate(
                    actual=self.y_test_label, pred=prediction, info=combination_descriptor)
                mlflow.log_metrics(logs)
                metrics.format_float()
                # TODO path bulunamadı hatası aldık
                print(
                    f'combination_artifact_uri:{combination_run.info.artifact_uri}')
                utils.compress_and_save_data(
                    metrics.__dict__, combination_run.info.artifact_uri, f'{self.date}_metrics.gzip')


class TFSingleModal:
    def __init__(self, date, descriptor, model_func, batch_size=128, epochs=100):
        self.date = date
        self.descriptor = descriptor
        self.model_func = model_func
        self.batch_size = batch_size
        self.epochs = epochs

    def set_data(self, train_idx_arr, val_idx_arr, train_data, train_label, test_data, test_label):
        self.train_idx_arr = train_idx_arr
        self.val_idx_arr = val_idx_arr
        self.train_data = train_data
        self.train_label = train_label
        self.test_data = test_data
        self.test_label = test_label

# https://github.com/mlflow/mlflow/blob/master/examples/tensorflow/train.py
    def predict(self):
        print(self.train_data.shape)

        # Failed to convert a NumPy array to a Tensor
        with mlflow.start_run(run_name=self.descriptor, description="***", nested=True) as run:
            models = dict()
            histories = dict()
            models_val_acc = dict()
            # with batch_metrics_logger(run_id) as metrics_logger:
            for i, (train_idx, val_idx) in enumerate(zip(self.train_idx_arr, self.val_idx_arr)):
                print(f"Validation {i}")

                with mlflow.start_run(run_name=f'Validation {i}', description='CV models', nested=True) as cv_fit:
                    model = self.model_func(self.train_data.shape[1])
                    models[f'validation_{i}'] = model
                    X_train_cv = self.train_data[train_idx]
                    y_train_cv = self.train_label[train_idx]
                    X_valid_cv = self.train_data[val_idx]
                    y_valid_cv = self.train_label[val_idx]

                    early_stopping = EarlyStopping(
                        monitor='val_loss', patience=10, verbose=0, mode='auto')
                    custom_callback = CustomCallback()
                    history = model.fit(X_train_cv, y_train_cv,
                                        batch_size=self.batch_size,
                                        epochs=self.epochs,
                                        validation_data=(
                                            X_valid_cv, y_valid_cv),
                                        callbacks=[early_stopping, custom_callback])
                    # histories[f'validation_{i}'] = history
                    models_val_acc[f'validation_{i}'] = history.history['val_accuracy'][-1]
                    # Saving each CV model

            best_model_key = max(models_val_acc, key=models_val_acc.get)
            best_model = models[best_model_key]
            best_model.evaluate(self.test_data, self.test_label,
                                callbacks=[custom_callback])
            pred = best_model.predict(self.test_data)

            logs, metrics = evaluate(
                actual=self.test_label, pred=pred, info=self.descriptor)
            metrics.format_float()
            mlflow.log_metrics(logs)
            mlflow.log_param('best_cv', best_model_key)
            signature = infer_signature(
                self.train_data,
                # generate_signature_output(model,X_valid_cv)
                # params=params,
            )

            mlflow.tensorflow.save_model(
                best_model,
                path=run.info.artifact_uri + '/model',
                signature=signature,
            )
            print(run.info.artifact_uri)
            onnx_model, _ = tf2onnx.convert.from_keras(
                best_model, input_signature=None, opset=13)
            onnx.save(onnx_model, run.info.artifact_uri +
                      '/model/model.onnx')
            utils.compress_and_save_data(
                metrics.__dict__, run.info.artifact_uri, f'{self.date}metrics.gzip')
            # mlflow.log_dict(metrics.__dict__, "metrics.json")

            # Plot Precision-Recall curves for each class and micro-average
            # fig = plt.figure()
            # plt.step(metrics.recall['micro_event'], metrics.precision['micro_event'],
            #          color='b', alpha=0.2, where='post')
            # plt.fill_between(
            #     metrics.recall["micro_event"], metrics.precision["micro_event"], step='post', alpha=0.2, color='b')

            # for i in range(pred.shape[1]):
            #     plt.step(metrics.recall[i], metrics.precision[i], where='post',
            #              label='Class {0} (AUC={1:0.2f})'.format(i, metrics.roc_aupr[i]))

            # plt.xlabel('Recall')
            # plt.ylabel('Precision')
            # plt.ylim([0.0, 1.05])
            # plt.xlim([0.0, 1.0])
            # plt.title(
            #     'Micro-average Precision-Recall curve: AUC={0:0.2f}'.format(metrics.roc_aupr["micro"]))
            # plt.legend(loc='best')
            # # plt.savefig(run.info.artifact_uri + '/auprc.png')
            # mlflow.log_figure(fig, 'auprc.png')

            # mlflow.log_model(
            #         model,
            #         artifact_path=run.info.artifact_uri + '/model',
            #         signature=signature,
            #     )
            # mlflow.log_artifact(run.info.artifact_uri + '/model')

            # mlflow.MlflowClient().log_artifact(run.info.run_id,
            #                                    run.info.artifact_uri, None)
        return pred


class CustomCallback(keras.callbacks.Callback):
    def on_train_begin(self, logs=None):
        keys = list(logs.keys())
        mlflow.log_param("train_begin_keys", keys)
        config = self.model.optimizer.get_config()
        for attribute in config:
            mlflow.log_param("opt_" + attribute, config[attribute])

        sum_list = []
        self.model.summary(print_fn=sum_list.append)
        summary = "\n".join(sum_list)
        mlflow.log_text(summary, artifact_file="model_summary.txt")

    def on_train_end(self, logs=None):
        print(logs)
        mlflow.log_metrics(logs)

    def on_epoch_begin(self, epoch, logs=None):
        keys = list(logs.keys())

    def on_epoch_end(self, epoch, logs=None):
        keys = list(logs.keys())

    def on_test_begin(self, logs=None):
        keys = list(logs.keys())

    def on_test_end(self, logs=None):
        mlflow.log_metrics(logs)
        print(logs)

    def on_predict_begin(self, logs=None):
        keys = list(logs.keys())

    def on_predict_end(self, logs=None):
        keys = list(logs.keys())
        mlflow.log_metrics(logs)

    def on_train_batch_begin(self, batch, logs=None):
        keys = list(logs.keys())

    def on_train_batch_end(self, batch, logs=None):
        keys = list(logs.keys())

    def on_test_batch_begin(self, batch, logs=None):
        keys = list(logs.keys())

    def on_test_batch_end(self, batch, logs=None):
        keys = list(logs.keys())

    def on_predict_batch_begin(self, batch, logs=None):
        keys = list(logs.keys())

    def on_predict_batch_end(self, batch, logs=None):
        keys = list(logs.keys())
    # def on_train_begin(self, logs=None):  # pylint: disable=unused-argument
    #     config = self.model.optimizer.get_config()
    #     for attribute in config:
    #         mlflow.log_param("opt_" + attribute, config[attribute])

    #     sum_list = []
    #     self.model.summary(print_fn=sum_list.append)
    #     summary = "\n".join(sum_list)
    #     mlflow.log_text(summary, artifact_file="model_summary.txt")

    # def on_epoch_end(self, epoch, logs=None):
    #     # NB: tf.Keras uses zero-indexing for epochs, while other TensorFlow Estimator
    #     # APIs (e.g., tf.Estimator) use one-indexing. Accordingly, the modular arithmetic
    #     # used here is slightly different from the arithmetic used in `_log_event`, which
    #     # provides  metric logging hooks for TensorFlow Estimator & other TensorFlow APIs
    #     if epoch % self.log_every_n_steps == 0:
    #         self.metrics_logger.record_metrics(logs, epoch)

    # def predict(self):
    #     model = self.model_func()
    #     # Failed to convert a NumPy array to a Tensor
    #     for i, (train_idx, val_idx) in enumerate(zip(self.train_idx_arr, self.val_idx_arr)):
    #         print(f"Validation {i}")
    #         X_train_cv = self.train_data[train_idx]
    #         y_train_cv = self.train_label[train_idx]
    #         X_valid_cv = self.train_data[val_idx]
    #         y_valid_cv = self.train_label[val_idx]

    #         early_stopping = EarlyStopping(
    #             monitor='val_loss', patience=10, verbose=0, mode='auto')
    #         model.fit(X_train_cv, y_train_cv, batch_size=128, epochs=20, validation_data=(X_valid_cv, y_valid_cv),
    #                   callbacks=[early_stopping])
    #     pred = model.predict(self.test_data)
    #     return pred



from .utils import create_folder_if_not_exists, utc_time_as_string,utc_time_as_string_simple_format, compress_and_save_data
from .zip_helper import ZipHelper
from .enums import UMLSCodeTypes, DrugBankTextDataTypes



import gzip
import json
import os

from datetime import datetime, timezone

from matplotlib import pyplot as plt


def create_folder_if_not_exists(path):
    if not os.path.exists(path):
        os.makedirs(path)


def utc_time_as_string():
    utc_datetime = datetime.now(timezone.utc)

    return datetime.strftime(utc_datetime, "%Y-%m-%dT%H:%M:%S.%f")[:-3]

def utc_time_as_string_simple_format():
    utc_datetime = datetime.now(timezone.utc)

    return datetime.strftime(utc_datetime, '%Y%m%d')

# https://gist.github.com/LouisAmon/4bd79b8ab80d3851601f3f9016300ac4


def compress_data(data):
    # Convert to JSON
    # json_data = json.dumps(data, indent=2)
    json_data = json.dumps(data, separators=(',', ":"))
    # Convert to bytes
    encoded = json_data.encode('UTF-8')
    # Compress
    compressed = gzip.compress(encoded)
    return compressed


def compress_and_save_data(data, path, file_name):
    compressed = compress_data(data)
    create_folder_if_not_exists(path)
    with gzip.open(path+f'/{file_name}', 'wb') as f:
        f.write(compressed)

def decompress(gzip_file):
    with gzip.open(gzip_file, 'r') as fin:        # 4. gzip
        json_bytes = fin.read()                      # 3. bytes (i.e. UTF-8)
    json_bytes = gzip.decompress(json_bytes)
    json_str = json_bytes.decode('UTF-8')            # 2. string (i.e. JSON)
    data = json.loads(json_str)
    return data


if __name__ == "__main__":
    # json_file = f'C:\\Users\\kivanc\\Downloads\\metrics.json'
    # file_data = open(json_file, "r", 1).read()
    # a = json.loads(file_data)  # store in json structure
    # # a = {'key1':1, 'key2':2}
    # compressed = compress_data(a)
    # with gzip.open('deneme.gzip', 'wb') as f:
    #     f.write(compressed)

    # with gzip.open('deneme.gzip', 'r') as fin:        # 4. gzip
    #     json_bytes = fin.read()                      # 3. bytes (i.e. UTF-8)
    # json_bytes = gzip.decompress(json_bytes)
    # json_str = json_bytes.decode('UTF-8')            # 2. string (i.e. JSON)
    # data = json.loads(json_str)
    # print(data)

    gzip_file = f'C:\\Users\\kivanc\\Downloads\\metrics (2).gzip'
    stored_file =  f'C:\\Users\\kivanc\\Downloads\\save.png'
    metrics = decompress(gzip_file)
    # print(metrics)

    # Plot Precision-Recall curves for each class and micro-average
    fig = plt.figure()
    plt.step(metrics['recall']['micro_event'], metrics['precision']['micro_event'],
             color='b', alpha=0.2, where='post')
    plt.fill_between(
        metrics['recall']["micro_event"], metrics['precision']["micro_event"], step='post', alpha=0.2, color='b')

    # for i in range(65):
    #     plt.step( metrics['recall'][str(i)],  metrics['precision'][str(i)], where='post',
    #              label='Class {0} (AUC={1:0.2f})'.format(i, metrics['roc_aupr'][str(i)]))

    plt.xlabel('Recall')
    plt.ylabel('Precision')
    plt.ylim([0.0, 1.05])
    plt.xlim([0.0, 1.0])
    plt.title(
        'Micro-average Precision-Recall curve: AUC={0:0.2f}'.format(metrics['roc_aupr']["micro"]))
    plt.legend(loc='best')
    plt.savefig(stored_file)
    # plt.show()

    import plotly.express as px
    import pandas as pd
    df = pd.DataFrame(dict(
    r=[1, 5, 2, 2, 3],
    theta=['processing cost','mechanical properties','chemical stability',
            'thermal stability', 'device integration']))
    fig = px.line_polar(df, r='r', theta='theta', line_close=True)
    fig.show()


from collections import defaultdict
import glob
import json
from pathlib import Path
import pathlib
from time import sleep
import pandas as pd

from tqdm import tqdm
import os
import requests
# from mmlrestclient as metamapliteclient
from enum import Enum
from ddi_fw.utils import create_folder_if_not_exists


# data = '''
# Lepirudin is a recombinant hirudin formed by 65 amino acids that acts as a highly specific and direct thrombin inhibitor.
# [L41539,L41569] Natural hirudin is an endogenous anticoagulant found in _Hirudo medicinalis_ leeches.
# [L41539] Lepirudin is produced in yeast cells and is identical to natural hirudin except for the absence of sulfate
# on the tyrosine residue at position 63 and the substitution of leucine for isoleucine at position 1 (N-terminal end).[A246609]

# Lepirudin is used as an anticoagulant in patients with heparin-induced thrombocytopenia (HIT),
# an immune reaction associated with a high risk of thromboembolic complications.[A3, L41539]
# HIT is caused by the expression of immunoglobulin G (IgG) antibodies that bind to the complex formed by heparin and platelet factor 4.
# This activates endothelial cells and platelets and enhances the formation of thrombi.
# [A246609] Bayer ceased the production of lepirudin (Refludan) effective May 31, 2012.[L41574]'''

# response = requests.post(url, data=data)

# print(response.content)

HERE = pathlib.Path(__file__).resolve().parent


class CTakesNER:
    def __init__(self, drugs_df = None,api_url= 'http://localhost:8080/ctakes-web-rest/service/analyze?pipeline=Default'
                  , output_path='ner-output/ctakes', ids=[],
                 columns=[]):
        self.drugs_df =  drugs_df
        self.api_url = api_url
        self.columns = columns
        self.ids = ids
        self.output_path = output_path

    def run(self, 
            run_for=[]):

        for column in self.columns:
            if not os.path.exists(self.output_path+"/"+column):
                os.makedirs(self.output_path+"/"+column)
        for column in self.columns:
            column_output_path = self.output_path+'/'+column
            if not column in run_for:
                continue
            # not include
            if self.ids:
                self.drugs_df = self.drugs_df[~self.drugs_df['drugbank_id'].isin(
                    self.ids)]
            for index, row in self.drugs_df.iterrows():
                drugbank_id = row['drugbank_id']
                data = row[column]
                if data is None or pd.isna(data) or (type(data) == str and len(data.strip()) == 0):  # or len(data) == 0:
                    with open(f'{column_output_path}/{drugbank_id}.json', 'w', encoding='utf-8') as f:
                        json.dump([], f, ensure_ascii=False, indent=4)
                    continue
                data = data.encode()
                response = requests.post(self.api_url, data=data)

                with open(f'{column_output_path}/{drugbank_id}.json', 'w', encoding='utf-8') as f:
                    try:
                        obj = json.loads(response.text)
                        json.dump(obj, f, ensure_ascii=False, indent=4)
                    except:
                        # print(f'{drugbank_id} is not parsable')
                        json.dump([], f, ensure_ascii=False, indent=4)
                        continue

                # if index % 10 == 0:
                #     sleep(10)

    def load(self, filename = None, group = True):
        file_path=  filename if filename else HERE.joinpath('output/ctakes/ctakes_ner.pkl')
        df = pd.read_pickle(file_path)

        if group:
            keys = list(df.columns.values)

            df['tui'] = [[]] * df.shape[0]
            df['cui'] = [[]] * df.shape[0]
            df['entities'] = [[]] * df.shape[0]

            tui_columns = [key for key in keys if key.startswith('tui')]
            cui_columns = [key for key in keys if key.startswith('cui')]
            entities_columns = [key for key in keys if key.startswith('entities')]
            #bunu tek bir eşitlikle çöz
            df['tui'] =  df[tui_columns].values.tolist()
            df['tui'] = df['tui'].apply(lambda items:{i for item in items for i in item})  

            df['cui'] =  df[cui_columns].values.tolist()
            df['cui'] = df['cui'].apply(lambda items:{i for item in items for i in item})  

            df['entities'] =  df[entities_columns].values.tolist()
            df['entities'] = df['entities'].apply(lambda items:{i for item in items for i in item})  

        return df

    def create_dataframe(self, override = False):  # dataframe_columns=[]
        filename='ctakes_ner.pkl'
        if not override and os.path.exists(self.output_path+"/" + filename):
            return self.load(self.output_path+"/" + filename)
        
        create_folder_if_not_exists(self.output_path+"/" + filename)
        dict_of_dict = defaultdict(dict)
        for column in self.columns:
            all_json_files = f'{self.output_path}/{column}/'+'*.json*'
            for filepath in tqdm(glob.glob(all_json_files)):
                with open(filepath, 'r', encoding="utf8") as f:
                    file_name = Path(f.name).stem
                    t = dict_of_dict[file_name]
                    data = json.load(f)
                    entities = []
                    cuis = []
                    tuis = []
                    if data is None or len(data) == 0:
                        t['drugbank_id'] = file_name
                        t[f'cui_{column}'] = []
                        t[f'tui_{column}']= []
                        t[f'entities_{column}'] = []
                        dict_of_dict[file_name] = t
                        continue
                    for key, value in data.items():
                        entities = [v['text'] for v in value]
                        cuis = [attr['cui']
                                for v in value for attr in v['conceptAttributes']]
                        tuis = [attr['tui']
                                for v in value for attr in v['conceptAttributes']]
                        # codingScheme
                    
                    if 'drugbank_id' not in t:
                        t['drugbank_id'] = file_name
                    t[f'cui_{column}'] = cuis
                    t[f'tui_{column}'] = tuis
                    t[f'entities_{column}'] = entities
                dict_of_dict[file_name] = t

        df = pd.DataFrame(dict_of_dict.values(),
                          #   orient='index',
                          #   columns=columns
                          )
        df.to_pickle(self.output_path+"/" + filename)
        # dataframe_columns.insert(0, 'drugbank_id')

        # new_columns = {columns[i]: dataframe_columns[i]
        #                for i in range(len(columns))}
        # df.rename(columns=new_columns, inplace=True)
        return df


# no module named 'mmlrestclient'
# class MMSLiteNER:

#     # https://ii.nlm.nih.gov/metamaplite/js/formControls.js

#     class Groups(Enum):

#         activities_group = ['acty', 'bhvr', 'dora',
#                             'evnt', 'gora', 'inbe', 'mcha', 'ocac', 'socb']

#         anatomy_group = ['anst', 'blor', 'bpoc', 'bsoj', 'bdsu',
#                          'bdsy', 'cell', 'celc', 'emst', 'ffas', 'tisu']

#         checmicals_and_drugs_group = ['aapp', 'antb', 'bacs', 'bodm', 'carb', 'chem', 'chvf', 'chvs',
#                                       'clnd', 'eico', 'elii', 'enzy', 'hops', 'horm', 'imft',
#                                       'irda', 'inch', 'lipd', 'nsba', 'nnon', 'orch', 'opco',
#                                       'phsu', 'rcpt', 'strd', 'vita']

#         concept_and_ideas_group = ['clas', 'cnce', 'ftcn', 'grpa', 'idcn', 'inpr', 'lang',
#                                    'qlco', 'rnlw', 'spco', 'tmco']

#         devices_group = ['drdd', 'medd', 'resd']

#         disorders_group = ['acab', 'anab', 'bact', 'comd', 'cgab', 'dsyn',
#                            'emod', 'fndg', 'inpo', 'mobd', 'patf', 'sosy']

#         # abbreviated disorders group, finding and congenital abnormality removed
#         disorders_abbrev_group = ['acab', 'anab', 'bact', 'cgab', 'dsyn',
#                                   'emod', 'inpo', 'mobd', 'patf', 'sosy']

#         genes_and_molecular_sequences = [
#             'amas', 'crbs', 'gngm', 'mosq', 'nusq']

#         geographic_areas = ['geoa']

#         living_being = ['aggp', 'amph', 'anim', 'arch', 'bact', 'bird', 'euka', 'fish',
#                         'fngs', 'grup', 'humn', 'mamm', 'orgm', 'podg',
#                         'plnt', 'popg', 'prog', 'rept', 'vtbt', 'virs']

#         objects = ['enty', 'food', 'mnob', 'sbst']

#         occupations = ['bmod', 'ocdi']

#         organizations = ['hcro', 'orgt', 'pros', 'shro']

#         phenomena = ['eehu' 'hcpp', 'lbtr', 'npop', 'phpr']

#         physiology = ['celf', 'clna', 'clnd']

#         procedures = ['diap', 'edac', 'hlca', 'lbpr', 'mbrt', 'resa', 'topp']

#     def __init__(self, drugs_df, input_path='drugbank/output', output_path='ner-output/metamaplite', ids=[],
#                  columns=[],
#                  included_groups: Groups = [],
#                  excluded_groups: Groups = [],
#                  ):

#         self.drugs_df = drugs_df
#         self.columns = columns
#         self.ids = ids
#         self.output_path = output_path
#         self.included_groups = set()
#         for i, g in enumerate(included_groups):
#             for v in g.value:
#                 self.included_groups.add(v)

#         self.excluded_groups = set()
#         for i, g in enumerate(excluded_groups):
#             for v in g.value:
#                 self.excluded_groups.add(v)

#         for column in columns:
#             if not os.path.exists(output_path+"/"+column):
#                 os.makedirs(output_path+"/"+column)

#     def run_ner(self):
#         # # url = 'https://ii-public1.nlm.nih.gov/metamaplite/rest/annotate'
#         base_url = 'https://ii.nlm.nih.gov/metamaplite/rest/annotate'
#         acceptfmt = 'text/plain'
#         for column in self.columns:
#             column_output_path = self.output_path+'/'+column

#             if self.ids:
#                 self.drugs_df = self.drugs_df[~self.drugs_df['drugbank_id'].isin(
#                     self.ids)]
#             for index, row in self.drugs_df.iterrows():
#                 drugbank_id = row['drugbank_id']
#                 input_text = row[column]
#                 params = [('inputtext', input_text), ('docformat', 'freetext'),
#                           ('resultformat', 'json'), ('sourceString', 'all'),
#                           ('semanticTypeString', 'all')]
#                 resp = metamapliteclient.handle_request(
#                     base_url, acceptfmt, params)

#                 with open(f'{column_output_path}/{drugbank_id}.json', 'w', encoding='utf-8') as f:
#                     obj = json.loads(resp.text)
#                     json.dump(obj, f, ensure_ascii=False, indent=4)

#                 if index % 10 == 0:
#                     sleep(10)

#     def __dict_of_semantic_types__(self, path):
#         m = dict()
#         with open(path, 'r', encoding='utf-8') as f:
#             data = f.read()
#             rows = data.split("\n")
#             for row in rows:
#                 if row != "":
#                     arr = row.split("|")
#                     m[arr[0]] = arr[1]
#         return m

#     def load(self, semantic_type_path: str, dataframe_columns=[]):
#         semantic_type_dict = self.__dict_of_semantic_types__(
#             semantic_type_path)

#         cui_dict = defaultdict(dict)
#         tui_dict = defaultdict(dict)
#         for column in self.columns:
#             all_json_files = f'{self.output_path}/{column}/'+'*.json*'
#             for filepath in tqdm(glob.glob(all_json_files)):
#                 with open(filepath, 'r', encoding="utf8") as f:
#                     file_name = Path(f.name).stem
#                     data = json.load(f)
#                     filtered_obj = [o for o in data if len(o['evlist']) == 1]
#                     # filtered_obj = [o for o in data if len(o['evlist']) == 1 and set(
#                     #     checmicals_and_drugs_group).intersection(set(o['evlist'][0]['conceptinfo']['semantictypes']))]

#                     if self.included_groups:
#                         evaluation = [o['evlist'][0]['conceptinfo'] for o in filtered_obj if len(o['evlist']) == 1
#                                       and
#                                       self.included_groups.intersection(
#                             set(o['evlist'][0]['conceptinfo']['semantictypes']))]
#                         # cuis = [o['evlist'][0]['conceptinfo']['cui'] for o in filtered_obj if len(o['evlist']) == 1
#                         #         and
#                         #         self.included_groups.intersection(
#                         #             set(o['evlist'][0]['conceptinfo']['semantictypes']))]
#                     elif self.excluded_groups:
#                         evaluation = cuis = [o['evlist'][0]['conceptinfo'] for o in filtered_obj if len(o['evlist']) == 1
#                                              and
#                                              not self.excluded_groups.intersection(
#                             set(o['evlist'][0]['conceptinfo']['semantictypes']))]
#                         # cuis = [o['evlist'][0]['conceptinfo']['cui'] for o in filtered_obj if len(o['evlist']) == 1
#                         #         and
#                         #         not self.excluded_groups.intersection(
#                         #             set(o['evlist'][0]['conceptinfo']['semantictypes']))]
#                     else:
#                         evaluation = [o['evlist'][0]['conceptinfo']
#                                       for o in filtered_obj if len(o['evlist']) == 1]
#                         # cuis = [o['evlist'][0]['conceptinfo']['cui']
#                         #         for o in filtered_obj if len(o['evlist']) == 1]

#                     # cuis = [o['evlist'][0]['conceptinfo']['cui'] for o in filtered_obj if len(o['evlist']) == 1 and set(
#                     #     checmicals_and_drugs_group).intersection(set(o['evlist'][0]['conceptinfo']['semantictypes']))]
#                     cuis = [ev['cui'] for ev in evaluation]
#                     semantic_types = [ev['semantictypes'] for ev in evaluation]
#                     tuis = [semantic_type_dict[s]
#                             for semantic_type in semantic_types for s in semantic_type]

#                     d = cui_dict[file_name]
#                     d['drugbank_id'] = file_name
#                     d[column] = set(cuis)

#                     t = tui_dict[file_name]
#                     t['drugbank_id'] = file_name
#                     t[column] = set(tuis)
#                     tui_dict[file_name] = t

#         columns = self.columns
#         columns.insert(0, 'drugbank_id')
#         df = pd.DataFrame(tui_dict.values(),
#                           #   orient='index',
#                           columns=columns
#                           )

#         dataframe_columns.insert(0, 'drugbank_id')

#         new_columns = {columns[i]: dataframe_columns[i]
#                        for i in range(len(columns))}
#         df.rename(columns=new_columns, inplace=True)
#         return df



'''
copied from https://github.com/YifanDengWHU/DDIMDL/blob/master/NLPProcess.py and reorganized
'''

# import stanfordnlp
# stanfordnlp.download("en")
import pandas as pd
import stanza
# stanza.download("en")

import numpy as np


class EventExtractor:
    def __init__(self, druglist, use_cache=True):
        self.druglist = druglist
        self.druglist2 = ['_'.join(d.replace('.', ' ').replace(
            ',', ' ').replace('-', ' ').split(' ')) for d in druglist]
        # self.events = events
        self.pipeline = stanza.Pipeline(use_gpu=True)
        self.cache = dict()

    def prepare_event_text(self, event):
        for ex, new in zip(self.druglist, self.druglist2):
            event = event.replace(ex, new)
        return event

    def extract_all(self, events):
        mechanisms = []
        actions = []
        drugA_list = []
        drugB_list = []
        for i in range(len(events)):
            mechanism, action, drugA, drugB = self.extract(events[i])
            mechanisms.append(mechanism)
            actions.append(action)
            drugA_list.append(drugA)
            drugB_list.append(drugB)
        return mechanisms, actions, drugA_list, drugB_list

    def extract(self, event):
        if event in self.cache:
            return self.cache[event]
        event = self.prepare_event_text(event)
        drugA = None
        drugB = None

        def addMechanism(node):
            if int(sonsNum[int(node-1)]) == 0:
                return
            else:
                for k in sons[node-1]:
                    if int(k) == 0:
                        break
                    if dependency[int(k - 1)].text == drugA or dependency[int(k - 1)].text == drugB:
                        continue
                    quene.append(int(k))
                    addMechanism(int(k))
            return quene

        doc = self.pipeline(event)
        dependency = []
        for j in range(len(doc.sentences[0].words)):
            dependency.append(doc.sentences[0].words[j])
        sons = np.zeros((len(dependency), len(dependency)))
        sonsNum = np.zeros(len(dependency))
        flag = False
        count = 0
        for j in dependency:
            # if j.dependency_relation=='root':
            if j.deprel == 'root':
                # root=int(j.index)
                root = int(j.id)
                action = j.lemma
            if j.text in self.druglist2:
                if count < 2:
                    if flag == True:
                        drugB = j.text
                        count += 1
                    else:
                        drugA = j.text
                        flag = True
                        count += 1
            sonsNum[j.head-1] += 1
            sons[j.head-1, int(sonsNum[j.head-1]-1)] = int(j.id)
        quene = []
        for j in range(int(sonsNum[root-1])):
            if dependency[int(sons[root-1, j]-1)].deprel == 'obj' or dependency[int(sons[root-1, j]-1)].deprel == 'nsubj:pass':
                quene.append(int(sons[root-1, j]))
                break
        quene = addMechanism(quene[0])
        quene.sort()

        mechanism = " ".join(dependency[j-1].text for j in quene)
        if mechanism == "the fluid retaining activities":
            mechanism = "the fluid"
        if mechanism == "atrioventricular blocking ( AV block )":
            mechanism = 'the atrioventricular blocking ( AV block ) activities increase'

        self.cache[event] = (mechanism, action,
                             drugA.replace('_', ' ') if drugA != None else '',
                             drugB.replace('_', ' ') if drugB != None else '')
        

        if drugA == '' or drugB == '':
            print(event)
 
        return mechanism, action, drugA.replace('_', ' ') if drugA != None else '',  drugB.replace('_', ' ') if drugB != None else ''


# drugs_pickle_path = 'drugbank/output/drugs.pkl'
# drugs_df = pd.read_pickle(drugs_pickle_path)

# drug_names = drugs_df['name'].to_list()


# drug_names = ['Lepirudin','Ursodeoxycholic acid']
# event_extractor = EventExtractor(
#     drug_names)

# mechanisms, actions, drugA_list, drugB_list = event_extractor.extract_all(
#     ['The risk or severity of bleeding and bruising can be increased when Lepirudin is combined with Ursodeoxycholic acid'])
# # mechanism, action, drugA, drugB = event_extractor.extract(
# #     'Bivalirudin may increase the anticoagulant activities of Bromfenac')


# print(mechanisms)



import numpy as np
from scipy.spatial.distance import pdist, squareform

# todo pd.unique kullan
def find_distinct_elements(frame):
    # y = set(pd.unique(frame))
    y = set()
    for x in frame:
        if x is not None:
            for k in x:
                #     if type(k) == list:
                #         for i in k:
                #             y.add(i)
                #     else:
                y.add(k)
    return y


def find_distinct_elements_count(frame):
    y = set()
    for x in frame:
        if x is not None:
            y.update(x)
    return len(y)


class SimilarityMatrixGenerator:
    def __init__(self):
        pass

    def create_jaccard_similarity_matrices_ex(self, array):
        jaccard_sim = 1 - pdist(array, metric='jaccard')
        jaccard_sim_matrix = squareform(jaccard_sim)
        return jaccard_sim_matrix

    # https://github.com/YifanDengWHU/DDIMDL/blob/master/DDIMDL.py , def Jaccard(matrix):
    def create_jaccard_similarity_matrices(self, matrix):
        matrix = np.mat(matrix)
        numerator = matrix * matrix.T
        denominator = np.ones(np.shape(matrix)) * matrix.T + \
            matrix * np.ones(np.shape(matrix.T)) - matrix * matrix.T
        matrix = numerator / denominator
        np.nan_to_num(matrix, nan=0.0)
        return matrix


class VectorGenerator:
    def __init__(self, df):
        self.df = df

    def generate_feature_vector(self, column):
        bit_vectors = []
        map = dict()
        idx = 0
        count = find_distinct_elements_count(self.df[column])
        print(f"find_distinct_elements_count bitti, boyut: {count}")
        for ind in self.df.index:
            e = self.df[column][ind]
            # vector = np.zeros(len(sorted_features))
            vector = np.zeros(count)
            if e is not None:
                for item in e:
                    if item in map:
                        vector[map[item]] = 1
                    else:
                        vector[idx]=1
                        map[item] = idx
                        idx += 1 
 
            bit_vectors.append(vector)
        print("array oluşturuldu")
        return np.array(bit_vectors)
    
    # def generate_feature_vector(self, column):
    #     bit_vectors = []
    #     distinct_feature = find_distinct_elements(self.df[column])
    #     sorted_features = sorted(distinct_feature)
    #     for ind in self.df.index:
    #         e = self.df[column][ind]
    #         vector = np.zeros(len(sorted_features))
    #         if e is not None:
    #             indexes = [i for i, x in enumerate(sorted_features) if x in e]
    #             np.put(vector, indexes, np.ones(len(indexes)))
    #         bit_vectors.append(vector)
    #     return bit_vectors

# bit_vectors ndarray olacak
    def generate_feature_vectors(self, columns):
        vectors = dict()
        for column in columns:
            bit_vectors = self.generate_feature_vector(column)
            vectors[column] = bit_vectors
        return vectors


# generate feature vector
# np.hstack

# https://www.datasciencelearner.com/how-to-create-an-array-of-bits-in-python/
#



# # https://caseolap.github.io/docs/drug/drugbank/
# #https://gist.github.com/rosherbal/56461421c69a8a7da775336c95fa62e0

import os
import zipfile
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import XMLParser, XMLPullParser
import pandas as pd
import xmlschema
import json as json
import sys
import unicodedata
import re

from ddi_fw.utils import ZipHelper


def slugify(value, allow_unicode=False):
    """
    Taken from https://github.com/django/django/blob/master/django/utils/text.py
    Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
    dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Convert to lowercase. Also strip leading and
    trailing whitespace, dashes, and underscores.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize('NFKC', value)
    else:
        value = unicodedata.normalize('NFKD', value).encode(
            'ascii', 'ignore').decode('ascii')
    value = re.sub(r'[^\w\s-]', '', value.lower())
    return re.sub(r'[-\s]+', '-', value).strip('-_')


def replace_key(key: str):
    if key.startswith('@'):
        key = key[1:]
    if key == '$':
        key = "value"
    elif '{http://www.drugbank.ca}' in key:
        key = key.replace('{http://www.drugbank.ca}', '')
    return key


def modify_keys(d):
    for k, v in d.copy().items():
        if isinstance(v, dict):
            d.pop(k)
            d[replace_key(k)] = v
            modify_keys(v)
        elif isinstance(v, list):
            d.pop(k)
            d[replace_key(k)] = v
            for i in v:
                if isinstance(i, list) or isinstance(i, dict):
                    modify_keys(i)
                # print(i)

        else:
            if k == "keyToChange":
                v = int(v)
            d.pop(k)
            d[replace_key(k)] = v
    return d


class DrugBankParser:
    def __init__(self, xsd_file='drugbank.xsd', zip_file='drugbank.zip', input_path='./drugbank'):

        # sys.path.insert(0,'/content/drive/My Drive/drugbank')
        # HERE = '/content/drive/My Drive/drugbank'
        HERE = input_path
        DRUGBANK_XSD = HERE + '/' + xsd_file
        DRUGBANK_ZIP = HERE + '/' + zip_file
        xsd = xmlschema.XMLSchema(DRUGBANK_XSD)
        self.drug_type_schema = xsd.complex_types[1]
        self.zf = zipfile.ZipFile(DRUGBANK_ZIP, 'r')

    def parse(self, save_path='./drugbank/drugs', override = False):
        if not override:
            print('No parsing process has been executed!!!')
            return

        elements = []
        k = 0

        for name in self.zf.namelist():
            f = self.zf.open(name)
            # tree = ET.parse(f)
            # root = tree.getroot()
            previous_element = None
            for event, element in ET.iterparse(f, events=('end',)):  # "end"
                # if k == 10:
                #     break
                if len(elements) == 0:
                    elements.append(element)
                elif len(elements) == 1:
                    elements.append(element)
                elif len(elements) == 2:
                    elements[0] = elements[1]
                    elements[1] = element
                if len(elements) == 2:
                    previous_element = elements[len(elements)-2]
                drug = None
                # previous_element = element.find("..")
                #
                if previous_element is not None and previous_element.tag == '{http://www.drugbank.ca}transporters' and event == 'end' and element.tag == "{http://www.drugbank.ca}drug":
                    drug = element
                    elements = []

                    # for child in element:
                    #     print(child.text)

                if drug is None:
                    continue

                name = drug.find("{http://www.drugbank.ca}name")

                d_name = None
                if name is not None:
                    d_name = name.text
                    line = name.text

                if d_name is None:
                    continue

                k = k + 1

                # print(d_name)

                # if lax is used we have to send d[0] as a parameter
                d = self.drug_type_schema.decode(drug, validation='strict')
                # pretty_dict = {replace_key(k): v for k, v in d[0].items()}
                pretty_dict = modify_keys(d)
                # for key, value in pretty_dict.items():
                #     print(key, '->', value)
                # file_name = slugify(d_name)

                from pathlib import Path

                Path(save_path).mkdir(parents=True, exist_ok=True)

                primary_id = [
                    id['value'] for id in pretty_dict["drugbank-id"] if id['primary'] == True][0]
                with open(f'{save_path}/{primary_id}.json', 'w', encoding='utf-8') as f:
                    json.dump(pretty_dict, f, ensure_ascii=False, indent=4)

        print("Done")

    def zip_files(self, chunk_size=1000, input_path='./drugbank/drugs', output_path='./drugbank/zips'):
        zip_helper = ZipHelper()
        zip_helper.zip(zip_prefix='drugs', input_path=input_path,
                       output_path=output_path, chunk_size=chunk_size)



# !pip install -U sentence-transformers

# from transformers import BertTokenizer,BertForPreTraining,BertModel
# from sentence_transformers import SentenceTransformer, util
import pandas as pd
import numpy as np
import nltk
from nltk import sent_tokenize
from tqdm import tqdm


nltk.download('punkt')

import os
def check_file_exists(path):
  return os.path.isdir(path)

def get_model_name_or_local_path(model_local_path, model_name):
  if check_file_exists(model_local_path):
    return model_local_path
  return model_name

import re
def process_text(text):
  text = re.sub("\[L\d*\]", "",text)
  text = text.replace("[","")
  text = text.replace("]","")
  return text



from collections import defaultdict
from functools import partial

# NOT modelden input size'ı anlama,
def create_embeddings(model, data, column, drop_column=True):
  # model._modules['1'].get_sentence_embedding_dimension()
  # shape = (1,model._modules['0'].get_word_embedding_dimension())
  shape = model._modules['0'].get_word_embedding_dimension()
  column_embeddings_dict = defaultdict(lambda: np.zeros(shape))
  for index, row in tqdm(data.iterrows()):
    # if index == 10:
    #   break
    text = data[column][index]
    # else'de zero
    if text == None or type(text) != str:
      embeddings = None
    else:
      sentences = sent_tokenize(text)
      embeddings = model.encode(sentences)

  #TODO benzer olan ilacın embedding değerini vererek dene
    if embeddings is None or len(embeddings) == 0: #embedding check none type
      sum_of_embeddings = np.zeros(shape)
    else:
      sum_of_embeddings = np.sum(embeddings, axis = 0)
    # column_embeddings_dict[row['id']] = sum_of_embeddings.reshape(1, -1) # 2d
    column_embeddings_dict[row['id']] = sum_of_embeddings
    # data.iloc[index][column+'_embedding']=sum_of_embeddings
  
  data[column+'_embedding'] = pd.Series(column_embeddings_dict.values())
  if(drop_column):
    data.drop([column], axis = 1, inplace = True)
  # data[column+'_embedding'] = [column_embeddings_dict[row['name']] for index, row in data.iterrows()]
  return column_embeddings_dict




import os
import pathlib
import sqlite3
from sqlite3 import Error
import pandas as pd

from ddi_fw.utils import ZipHelper

from ..core import BaseDataset
# from ..db_utils import create_connection, select_all_drugs_as_dataframe, select_events_with_category

HERE = pathlib.Path(__file__).resolve().parent


class MDFSADDIDataset(BaseDataset):
    def __init__(self, chemical_property_columns=['enzyme',
                                                  'target',
                                                  'smile'],
                 embedding_columns=[],
                 ner_columns=[],
                 **kwargs):

        super().__init__(chemical_property_columns, embedding_columns,
                         ner_columns, **kwargs)

        db_zip_path = HERE.joinpath('mdf-sa-ddi.zip')
        db_path = HERE.joinpath('mdf-sa-ddi.db')
        if not os.path.exists(db_zip_path):
            self.__to_db__(db_path)
        else:
            ZipHelper().extract(
                input_path=str(HERE), output_path=str(HERE))
            conn = create_connection(db_path)
            self.drugs_df = select_all_drugs_as_dataframe(conn)
            self.ddis_df = select_all_events_as_dataframe(conn)
        kwargs = {'index_path': str(HERE.joinpath('indexes'))}

        self.index_path = kwargs.get('index_path')

    def __to_db__(self, db_path):
        conn = create_connection(db_path)
        drugs_path = HERE.joinpath('drug_information_del_noDDIxiaoyu50.csv')
        ddis_path = HERE.joinpath('df_extraction_cleanxiaoyu50.csv')
        self.drugs_df = pd.read_csv(drugs_path)
        self.ddis_df = pd.read_csv(ddis_path)
        self.drugs_df.drop(columns="Unnamed: 0", inplace=True)
        self.ddis_df.drop(columns="Unnamed: 0", inplace=True)

        self.ddis_df.rename(
            columns={"drugA": "name1", "drugB": "name2"}, inplace=True)
        self.ddis_df['event_category'] = self.ddis_df['mechanism'] + \
            ' ' + self.ddis_df['action']

        reverse_ddis_df = pd.DataFrame()
        reverse_ddis_df['id1'] = self.ddis_df['id2']
        reverse_ddis_df['name1'] = self.ddis_df['name2']
        reverse_ddis_df['id2'] = self.ddis_df['id1']
        reverse_ddis_df['name2'] = self.ddis_df['name1']
        reverse_ddis_df['event_category'] = self.ddis_df['event_category']

        self.ddis_df = pd.concat(
            [self.ddis_df, reverse_ddis_df], ignore_index=True)

        drug_name_id_pairs = {}
        for idx, row in self.drugs_df.iterrows():
            drug_name_id_pairs[row['name']] = row['id']

        # id1,id2

        def lambda_fnc1(column):
            return drug_name_id_pairs[column]
        # def lambda_fnc2(row):
        #     x  = self.drugs_df[self.drugs_df['name'] == row['name2']]
        #     return x['id']

        self.ddis_df['id1'] = self.ddis_df['name1'].apply(
            lambda_fnc1)  # , axis=1
        self.ddis_df['id2'] = self.ddis_df['name2'].apply(
            lambda_fnc1)  # , axis=1
        self.drugs_df.to_sql('drug', conn, if_exists='replace', index=False)
        self.ddis_df.to_sql('event', conn, if_exists='replace', index=False)
        ZipHelper().zip_single_file(
            file_path=db_path, output_path=HERE, name='mdf-sa-ddi')


def create_connection(db_file=r"mdf-sa-ddi.db"):
    """ create a database connection to the SQLite database
        specified by db_file
    :param db_file: database file
    :return: Connection object or None
    """
    conn = None
    try:
        conn = sqlite3.connect(db_file)
    except Error as e:
        print(e)

    return conn


def select_all_drugs(conn):
    cur = conn.cursor()
    cur.execute(
        '''select "index", id, name, target, enzyme, smile from drug''')
    rows = cur.fetchall()
    return rows


def select_all_drugs_as_dataframe(conn):
    headers = ['index', 'id', 'name', 'target', 'enzyme', 'smile']
    rows = select_all_drugs(conn)
    df = pd.DataFrame(columns=headers, data=rows)
    df['enzyme'] = df['enzyme'].apply(lambda x: x.split('|'))
    df['target'] = df['target'].apply(lambda x: x.split('|'))
    df['smile'] = df['smile'].apply(lambda x: x.split('|'))
    return df


def select_all_events(conn):
    """
    Query all rows in the event table
    :param conn: the Connection object
    :return:
    """
    cur = conn.cursor()
    cur.execute('''
                select event."index", id1, name1, id2, name2, mechanism, action, event_category from event
                ''')

    rows = cur.fetchall()
    return rows


def select_all_events_as_dataframe(conn):
    headers = ["index", "id1", "name1", "id2",
               "name2", "mechanism", "action", "event_category"]
    rows = select_all_events(conn)
    return pd.DataFrame(columns=headers, data=rows)



import zipfile as z
import os
from os.path import basename
from collections import defaultdict
import math


class ZipHelper:
    def __init__(self):
        pass

    def zip_single_file(self, name, file_path, output_path):
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        with z.ZipFile(f'{output_path}/{name}.zip', 'w', compression=z.ZIP_LZMA, compresslevel=z.ZIP_LZMA) as zipObj:
            zipObj.write(file_path, basename(file_path))

    def zip(self, zip_prefix, input_path, output_path, chunk_size):
        files_paths = [input_path+'/' + p for p in os.listdir(input_path)]
        count_of_chunks = math.ceil(len(files_paths) / chunk_size)
        zero_padding_length = len(str(int(count_of_chunks))) + 2

        if not os.path.exists(output_path):
            os.makedirs(output_path)

        part = 1
        i = 0
        zip_dict = defaultdict(list)
        for filePath in files_paths:
            padded_part = f'{part}'.zfill(zero_padding_length)
            key = f'{zip_prefix}.{padded_part}'
            zip_dict[key].append(filePath)
            i += 1
            if i % chunk_size == 0:
                i = 0
                part += 1

        for key, value in zip_dict.items():
            with z.ZipFile(f'{output_path}/{key}.zip', 'w', compression=z.ZIP_LZMA, compresslevel=z.ZIP_LZMA) as zipObj:
                for file_path in value:
                    zipObj.write(file_path, basename(file_path))

    def extract(self, input_path, output_path):
        files_paths = [input_path+'/' + p for p in os.listdir(input_path)]
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        for file_path in files_paths:
            if file_path.endswith('zip'):
                with z.ZipFile(file_path, 'r') as z1:
                    z1.extractall(path=output_path)
                    print(f'{file_path} has been extracted')


# if __name__ == "__main__":
#     helper = ZipHelper()
    # helper.zip(zip_prefix='drugs', input_path='drugbank/drugs',
    #            output_path='drugbank/drugs-zips', chunk_size=1000)
    # helper.extract(input_path='drugbank/drugs-zips',
    #                output_path='drugbank/drugs-extracted')
    # path = ''
    # import pandas as pd
    # d = {'col1': [1, 2], 'col2': [3, 4]}
    # df = pd.DataFrame(data=d)
    # df.to_pickle('test/dataframe.pickle')
    # helper.zip_single_file(file_path='test/dataframe.pickle',output_path='test/output', name='zip')
    # helper.extract(input_path='test/output', output_path='test/output')


import pathlib

import pandas as pd
from .. import BaseDataset
from ..db_utils import create_connection

HERE = pathlib.Path(__file__).resolve().parent


class DDIMDLDataset(BaseDataset):
    def __init__(self,  chemical_property_columns=['enzyme',
                                                   'target',
                                                   'pathway',
                                                   'smile'],
                 embedding_columns=[],
                 ner_columns=[],
                 **kwargs):

        super().__init__(chemical_property_columns, embedding_columns,
                         ner_columns, **kwargs)

        # kwargs = {'index_path': str(HERE.joinpath('indexes'))}
        kwargs['index_path']= str(HERE.joinpath('indexes'))

        db = HERE.joinpath('event.db')
        conn = create_connection(db)
        print("db prep")
        self.drugs_df = self.__select_all_drugs_as_dataframe__(conn)
        self.ddis_df = self.__select_all_events__(conn)
        print("db bitti")
        self.index_path = kwargs.get('index_path')

    def __select_all_drugs_as_dataframe__(self, conn):
        headers = ['index', 'id', 'name',
                   'target', 'enzyme', 'pathway', 'smile']
        cur = conn.cursor()
        cur.execute(
            '''select "index", id, name, target, enzyme, pathway, smile from drug''')
        rows = cur.fetchall()
        df = pd.DataFrame(columns=headers, data=rows)
        df['enzyme'] = df['enzyme'].apply(lambda x: x.split('|'))
        df['target'] = df['target'].apply(lambda x: x.split('|'))
        df['pathway'] = df['pathway'].apply(lambda x: x.split('|'))
        df['smile'] = df['smile'].apply(lambda x: x.split('|'))
        return df

    def __select_all_events__(self, conn):
        """
        Query all rows in the event table
        :param conn: the Connection object
        :return:
        """
        cur = conn.cursor()
        cur.execute('''
                select ex."index", d1.id, d1.name, d2.id, d2.name,  mechanism || ' ' ||action from extraction ex
                join drug d1 on  d1.name = ex.drugA
                join drug d2 on  d2.name = ex.drugB
        ''')

        rows = cur.fetchall()

        headers = ["index", "id1", "name1", "id2", "name2", "event_category"]
        return pd.DataFrame(columns=headers, data=rows)



from .core import BaseDataset
from .ddi_mdl.base import DDIMDLDataset 
from .mdf_sa_ddi.base import MDFSADDIDataset
from .custom import CustomDataset
from .embedding_generator import create_embeddings
from .embedding_generator_new import EmbeddingGenerator,PretrainedEmbeddingGenerator,SBertEmbeddingGenerator,LLMEmbeddingGenerator,create_embeddings_new
from .idf_helper import IDF
from .feature_vector_generation import SimilarityMatrixGenerator, VectorGenerator

__all__  = ['BaseDataset','DDIMDLDataset','MDFSADDIDataset']





from setuptools import setup, find_packages
setup(
    name='ddi_fw',
    version='0.0.6',
    author='Kıvanç Bayraktar',
    author_email='bayraktarkivanc@gmail.com',
    description='Do not use :)',
    # packages=find_packages(),
    packages=['ddi_fw/datasets', 'ddi_fw/datasets/mdf_sa_ddi', 'ddi_fw/datasets/ddi_mdl',
              'ddi_fw/datasets/custom', 'ddi_fw/drugbank',  'ddi_fw/experiments', 'ddi_fw/ner', 'ddi_fw/utils'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)




from sqlite3 import Error
import sqlite3
import pandas as pd
import numpy as np


def create_connection(db_file=r"./event.db"):
    """ create a database connection to the SQLite database
        specified by db_file
    :param db_file: database file
    :return: Connection object or None
    """
    conn = None
    try:
        conn = sqlite3.connect(db_file)
    except Error as e:
        print(e)

    return conn


def select_all_drugs(conn):
    cur = conn.cursor()
    cur.execute(
        '''select "index", id, name, target, enzyme, pathway, smile from drug''')
    rows = cur.fetchall()
    return rows


def select_all_drugs_as_dataframe(conn):
    headers = ['index','id', 'name', 'target', 'enzyme', 'pathway', 'smile']
    rows = select_all_drugs(conn)
    df = pd.DataFrame(columns=headers, data=rows)
    df['enzyme'] = df['enzyme'].apply(lambda x: x.split('|'))
    df['target'] = df['target'].apply(lambda x: x.split('|'))
    df['pathway'] = df['pathway'].apply(lambda x: x.split('|'))
    df['smile'] = df['smile'].apply(lambda x: x.split('|'))
    return df


def select_all_events(conn):
    """
    Query all rows in the event table
    :param conn: the Connection object
    :return:
    """
    cur = conn.cursor()
    cur.execute("select * from event")

    rows = cur.fetchall()
    return rows


def select_all_events_as_dataframe(conn):
    headers = ["index", "id1", "name1", "id2", "name2", "event_category"]
    rows = select_all_events(conn)
    return pd.DataFrame(columns=headers, data=rows)


def select_events_with_category(conn):
    sql = '''select id1, name1, id2, name2, mechanism || ' ' ||action from event ev
            join extraction ex
            on ev.name1 = ex.drugA and ev.name2 = ex.drugB
            union 
            select id1, name1, id2, name2, mechanism || ' ' ||action from event ev
            join extraction ex
            on ev.name1 = ex.drugB and ev.name2 = ex.drugA
            '''
    cur = conn.cursor()
    cur.execute(sql)

    rows = cur.fetchall()

    headers = ['id1', 'name1', 'id2', 'name2', 'event_category']
    return pd.DataFrame(columns=headers, data=rows)


def select_all_interactions_tuple_as_dataframe(conn):
    cur = conn.cursor()
    cur.execute("select id1, id2 from event")
    rows = cur.fetchall()
    headers = ['id1', 'id2']

    return pd.DataFrame(columns=headers, data=rows)


def select_ddi_pairs(conn):
    cur = conn.cursor()
    cur.execute('''
        select d1.[index] as Drug1Index, d2.[index] as Drug2Index, 1 from event e 
        join drug d1 on e.id1 = d1.id 
        join drug d2 on e.id2 = d2.id 
    ''')
    rows = cur.fetchall()
    return rows


def select_ddi_pairs_as_dataframe(conn):
    headers = ["Drug1Index", "Drug2Index", "Interaction"]
    rows = select_ddi_pairs(conn)
    return pd.DataFrame(columns=headers, data=rows)


def get_interactions(conn):
    cur = conn.cursor()
    cur.execute('''
        select
            drug_1_id,
            drug_1,
            drug_2_id,
            drug_2,
            mechanism_action, 
            interaction, 
            masked_interaction
        from _Interactions
                ''')

    rows = cur.fetchall()

    headers = ['id1', 'name1', 'id2', 'name2',
               'event_category', 'interaction', 'masked_interaction']
    df = pd.DataFrame(columns=headers, data=rows)
    return df


def get_extended_version(conn):
    cur = conn.cursor()
    cur.execute('''
        select 
        _Drugs."index",
        drugbank_id,
        _Drugs.name,
        description,
        synthesis_reference,
        indication,
        pharmacodynamics,
        mechanism_of_action,
        toxicity,
        metabolism,
        absorption,
        half_life,
        protein_binding,
        route_of_elimination,
        volume_of_distribution,
        clearance,
        smiles,
        smiles_morgan_fingerprint,
        enzymes_polypeptides,
        targets_polypeptides
                
        from drug 
        join _Drugs on drug.id = _Drugs.drugbank_id  
        where 
                targets_polypeptides is not null and 
                enzymes_polypeptides is not null and 
                smiles_morgan_fingerprint is not null
''')
    # pathway is absent

    rows = cur.fetchall()
    headers = ['index', 'id', 'name', 'description', 'synthesis_reference', 'indication', 'pharmacodynamics', 'mechanism_of_action', 'toxicity', 'metabolism', 'absorption', 'half_life',
               'protein_binding', 'route_of_elimination', 'volume_of_distribution', 'clearance', 'smiles_notation', 'smile', 'enzyme', 'target']
    df = pd.DataFrame(columns=headers, data=rows)
    df['smile'] = df['smile'].apply(lambda x:
                                    np.fromstring(
                                        x.replace(
                                            '\n', '')
                                        .replace('[', '')
                                        .replace(']', '')
                                        .replace('  ', ' '), sep=','))
    df['enzyme'] = df['enzyme'].apply(
        lambda x: x.split('|'))
    df['target'] = df['target'].apply(
        lambda x: x.split('|'))
    return df


# SELECT
#  CASE
#  WHEN masked_interaction like '%'+drug_1+'%' THEN drug_1
#  WHEN masked_interaction like '%'+drug_2+'%' THEN drug_2
#  Else drug_2
#  END AS Absent,

# drug_1, drug_2,
# masked_interaction

# from _Interactions
# where LENGTH(masked_interaction) = LENGTH(REPLACE(masked_interaction, 'DRUG', ''))
# or LENGTH(masked_interaction) = LENGTH(REPLACE(masked_interaction, 'DRUG', '')) + 4

if __name__ == "__main__":
    conn = create_connection(r"./event-extended.db")
    extended_version_df = get_extended_version(conn)

    df = select_all_events_as_dataframe(conn)
    print(df.head())

    events_with_category_df = select_events_with_category(conn)
    print(events_with_category_df.head())

    u = events_with_category_df['event_category'].unique()
    print(len(u))



import pandas as pd
import os
import json
import glob
from tqdm import tqdm

import csv

from rdkit import Chem
from rdkit.Chem import AllChem
import numpy as np
from ddi_fw.drugbank.event_extractor import EventExtractor

from zip_helper import ZipHelper
# from event_extractor import EventExtractor


def multiline_to_singleline(multiline):
    if multiline is None:
        return ""
    return " ".join(line.strip() for line in multiline.splitlines())

# targets -> target -> polypeptide
# enzymes -> enzyme -> polypeptide
# pathways from KEGG, KEGG ID is obtained from ddi_fw.drugbank
# https://www.genome.jp/dbget-bin/www_bget?drug:D03136
# https://www.kegg.jp/entry/D03136


class DrugBankProcessor():

    def mask_interaction(self, drug_1, drug_2, interaction):
        return interaction.replace(
            drug_1, "DRUG").replace(drug_2, "DRUG")

    def extract_zip_files(self, input_path='zips', output_path='drugs', override=False):
        if override:
            zip_helper = ZipHelper()
            zip_helper.extract(input_path=input_path, output_path=output_path)

    def process(self, input_path='drugs', output_path='output', zip_outputs=True):
        if not os.path.exists(output_path):
            os.makedirs(output_path)

        drugs_pickle_path = output_path+'/drugs.pkl'
        drugs_csv_path = output_path+'/drugs.gzip'
        ddi_pickle_path = output_path + '/ddi.pkl'
        ddi_csv_path = output_path + '/ddi.gzip'

        if not os.path.exists(drugs_pickle_path) or not os.path.exists(ddi_pickle_path):
            drug_rows = []
            all_ddis = []
            all_json_files = input_path+'/*.json*'

            for filepath in tqdm(glob.glob(all_json_files)):
                with open(filepath, 'r', encoding="utf8") as f:

                    data = json.load(f)

                    # if data['drug-interactions'] is None:
                    if False:
                        continue
                    else:
                        drug_1 = data['name']
                        drug_1_id = [d['value']
                                     for d in data['drugbank-id'] if d['primary'] == True][0]
                        description = multiline_to_singleline(
                            data['description'])
                        if data['drug-interactions'] is not None:
                            drug_interactions = [
                                interaction for interaction in data['drug-interactions']['drug-interaction']]
                            ddis = [(drug_1, interaction['name'], interaction['description'])
                                    for interaction in data['drug-interactions']['drug-interaction']]

                            ddi_dict = [{
                                'drug_1_id': drug_1_id,
                                'drug_1': drug_1,
                                'drug_2_id': interaction['drugbank-id']['value'],
                                'drug_2': interaction['name'],
                                'interaction': interaction['description'],
                                'masked_interaction': self.mask_interaction(drug_1, interaction['name'], interaction['description'])}
                                for interaction in data['drug-interactions']['drug-interaction']]
                            all_ddis.extend(ddi_dict)

                        synthesis_reference = data['synthesis-reference']
                        indication = multiline_to_singleline(
                            data['indication'])
                        pharmacodynamics = multiline_to_singleline(
                            data['pharmacodynamics'])
                        mechanism_of_action = multiline_to_singleline(
                            data['mechanism-of-action'])
                        toxicity = multiline_to_singleline(data['toxicity'])
                        metabolism = multiline_to_singleline(
                            data['metabolism'])
                        absorption = multiline_to_singleline(
                            data['absorption'])
                        half_life = multiline_to_singleline(data['half-life'])
                        protein_binding = multiline_to_singleline(
                            data['protein-binding'])
                        route_of_elimination = multiline_to_singleline(
                            data['route-of-elimination'])
                        volume_of_distribution = multiline_to_singleline(
                            data['volume-of-distribution'])
                        clearance = multiline_to_singleline(data['clearance'])

                        food_interactions = data['food-interactions']
                        sequences = data['sequences'] if "sequences" in data else None

                        external_identifiers = data['external-identifiers'] if "external-identifiers" in data else None
                        experimental_properties = data['experimental-properties'] if "experimental-properties" in data else None
                        calculated_properties = data['calculated-properties'] if "calculated-properties" in data else None

                        enzymes_polypeptides = None
                        targets_polypeptides = None

                        # targets = data['targets'] if "targets" in data else None
                        if data['targets'] is not None:
                            # targets_polypeptides = [p['id'] for d in data['targets']['target'] for p in d['polypeptide'] if 'polypeptide' in d ]
                            targets_polypeptides = [
                                p['id'] for d in data['targets']['target'] if 'polypeptide' in d for p in d['polypeptide']]

                        if data['enzymes'] is not None:
                            # enzymes_polypeptides = [p['id'] for d in data['enzymes']['enzyme'] for p in d['polypeptide'] if 'polypeptide' in d]
                            enzymes_polypeptides = [
                                p['id'] for d in data['enzymes']['enzyme'] if 'polypeptide' in d for p in d['polypeptide']]

                        if external_identifiers is not None:
                            external_identifiers_dict = dict(
                                [(p['resource'], p['identifier']) for p in external_identifiers['external-identifier']])

# add note column
                        smiles = None
                        morgan_hashed = None
                        if calculated_properties is not None:
                            calculated_properties_dict = dict(
                                [(p['kind'], p['value']) for p in calculated_properties['property']])
                            smiles = calculated_properties_dict['SMILES'] if 'SMILES' in calculated_properties_dict else None
                            if smiles is not None:
                                try:
                                    mol = Chem.MolFromSmiles(smiles)
                                    morgan_hashed = AllChem.GetMorganFingerprintAsBitVect(
                                        mol, 2, nBits=881).ToList()
                                except:
                                    print("An exception occurred")
                        if morgan_hashed is None:
                            morgan_hashed = np.zeros(881)

                        # k = [p[k]  for p in calculated_properties['property'] for k in p.keys() if k =='SMILES']
                        # external_identifiers['external-identifier']
                        # experimental_properties['property']

                        row = {'drugbank_id': drug_1_id,
                               'name': drug_1,
                               'description': description,
                               'synthesis_reference': synthesis_reference,
                               'indication': indication,
                               'pharmacodynamics': pharmacodynamics,
                               'mechanism_of_action': mechanism_of_action,
                               'toxicity': toxicity,
                               'metabolism': metabolism,
                               'absorption': absorption,
                               'half_life': half_life,
                               'protein_binding': protein_binding,
                               'route_of_elimination': route_of_elimination,
                               'volume_of_distribution': volume_of_distribution,
                               'clearance': clearance,
                               'smiles': smiles,
                               'smiles_morgan_fingerprint': morgan_hashed,
                               'enzymes_polypeptides': enzymes_polypeptides,
                               'targets_polypeptides': targets_polypeptides,
                               'external_identifiers': external_identifiers_dict
                               }
                        drug_rows.append(row)

                    # if len(drug_rows) == 10:
                    #     break
            # print(smiles_count)
            print(f"Size of drugs {len(drug_rows)}")
            print(f"Size of DDIs {len(all_ddis)}")
            np.set_printoptions(threshold=np.inf)

            # drug_names = [row['name'] for row in drug_rows]
            drug_names = ['DRUG']
            event_extractor = EventExtractor(drug_names)

            replace_dict = {'MYO-029': 'Stamulumab'}
            for ddi in tqdm(all_ddis):
                for key, value in replace_dict.items():
                    ddi['masked_interaction'] = ddi['masked_interaction'].replace(
                        key, value)
            #     interaction = ddi['interaction']
            #     mechanism, action, drugA, drugB = event_extractor.extract(interaction)
            #     ddi['mechanism'] = mechanism
            #     ddi['action'] = action

            self.drugs_df = pd.DataFrame(drug_rows)
            self.drugs_df.to_pickle(drugs_pickle_path)
            self.drugs_df.to_csv(
                drugs_csv_path,  index=False, compression='gzip')

            # print('mechanism_action calculation')
            self.ddis_df = pd.DataFrame(all_ddis)

            count = [0]

            def fnc2(interaction, count):
                count[0] = count[0] + 1
                if count[0] % 1000 == 0:
                    print(f'{count[0]}/{len(all_ddis)}')
                mechanism, action, drugA, drugB = event_extractor.extract(
                    interaction)
                return mechanism+'__' + action

            # self.ddis_df['mechanism_action'] = self.ddis_df['interaction'].apply(lambda x: fnc2(x))
            # tqdm.pandas()
            self.ddis_df['mechanism_action'] = self.ddis_df['masked_interaction'].apply(
                fnc2, args=(count,))

            self.ddis_df.to_csv(ddi_csv_path,  index=False, compression='gzip')
            self.ddis_df.to_pickle(ddi_pickle_path)

            if zip_outputs:
                zip_helper = ZipHelper()
                zip_helper.zip_single_file(
                    file_path=drugs_pickle_path, output_path=output_path+'/zips', name='drugs-pickle')
                zip_helper.zip_single_file(
                    file_path=ddi_pickle_path, output_path=output_path+'/zips', name='ddi-pickle')

        else:
            print('Output path has processed data, load function is called')
            self.load(output_path)

    def load(self, path):
        drugs_pickle_path = path+'/drugs.pkl'
        ddi_pickle_path = path+'/ddi.pkl'
        if os.path.exists(drugs_pickle_path) and os.path.exists(ddi_pickle_path):
            self.drugs_df = pd.read_pickle(drugs_pickle_path)
            self.ddis_df = pd.read_pickle(ddi_pickle_path)
        else:
            print('One of given paths could not found')

    def load_from_csv(self, path):
        drugs_csv_path = path+'/drugs.gzip'
        ddi_csv_path = path+'/ddi.gzip'
        if os.path.exists(drugs_csv_path) and os.path.exists(ddi_csv_path):
            self.drugs_df = pd.read_csv(drugs_csv_path, compression='gzip')
            self.ddis_df = pd.read_csv(ddi_csv_path, compression='gzip')
        else:
            print('One of given paths could not found')

    def load2(self, path):
        drugs_pickle_path = path+'/drugs.pkl'
        ddi_csv_path = path+'/ddi.gzip'
        if os.path.exists(drugs_pickle_path) and os.path.exists(ddi_csv_path):
            self.drugs_df = pd.read_pickle(drugs_pickle_path)
            self.ddis_df = pd.read_csv(ddi_csv_path, compression='gzip')
        else:
            print('One of given paths could not found')

    def drugs_as_dataframe(self):
        return self.drugs_df

    def filtered_drugs_as_dataframe(self, drug_ids):
        return self.drugs_df[self.drugs_df['drugbank_id'].isin(drug_ids)]

    def ddis_as_dataframe(self):
        return self.ddis_df

    def filtered_ddis(self, drugs):
        ddis_df = self.ddis_df.copy()
        return ddis_df[(ddis_df['drug_1'] in drugs) & (
            ddis_df['drug_2'] in drugs)]



# !pip install -U sentence-transformers

# from transformers import BertTokenizer,BertForPreTraining,BertModel
# from sentence_transformers import SentenceTransformer, util
import pandas as pd
import numpy as np
from nltk import sent_tokenize
import torch
from tqdm import tqdm


from collections import defaultdict
from functools import partial
from abc import ABC, abstractmethod
from transformers import AutoModel, AutoTokenizer
from sentence_transformers import SentenceTransformer, util


class EmbeddingGenerator(ABC):

    def __init__(self):
        self.shape = None

    @abstractmethod
    def generate(self, text):
        pass

# https://github.com/huggingface/transformers/issues/1791
class PretrainedEmbeddingGenerator(EmbeddingGenerator):
    def __init__(self, model_name, split_text=True):
        self.model_name = model_name
        self.model = AutoModel.from_pretrained(model_name)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.shape = self.model.get_input_embeddings().weight.shape
        self.split_text = split_text

    def generate(self, text):
        if self.split_text:
            sentences = sent_tokenize(text)
            output_embeddings = None
            for sentence in sentences:
                input_ids  = self.tokenizer.encode(sentence, return_tensors='pt', padding=True)
                if output_embeddings == None:
                    output_embeddings = self.model(input_ids).last_hidden_state.mean(dim=1)
                else:
                    output_embeddings += self.model(input_ids).last_hidden_state.mean(dim=1)
            if output_embeddings == None:
                output_embeddings = torch.empty((1,self.model.get_input_embeddings().weight.shape[1]))
        else:
            encoded_input = self.tokenizer(text, return_tensors='pt')
            input_ids = self.tokenizer.encode(text, add_special_tokens=True, max_length=self.tokenizer.model_max_length, return_tensors='pt')
            # input_ids  = encoded_input.input_ids[:self.tokenizer.model_max_length]
            output_embeddings = self.model(input_ids)
            # output_embeddings = self.model(**encoded_input)
            # sentence embedding
            output_embeddings = output_embeddings.last_hidden_state.mean(dim=1)
        return torch.flatten(output_embeddings).detach().numpy()


class LLMEmbeddingGenerator(EmbeddingGenerator):
    pass


class SBertEmbeddingGenerator(PretrainedEmbeddingGenerator):
    def __init__(self, model_name, split_text=True):
        self.model = SentenceTransformer(model_name)
        self.shape = self.model._modules['0'].get_word_embedding_dimension()
        self.split_text = split_text

    def generate(self, text):
        if text == None or type(text) != str:
            embeddings = None
        else:
            if self.split_text:
                sentences = sent_tokenize(text)
                embeddings = self.model.encode(sentences)
            else:
                embeddings = self.model.encode(text)
        return embeddings


# NOT modelden input size'ı anlama,
def create_embeddings_new(generator: EmbeddingGenerator, data, column, drop_column=True):
    column_embeddings_dict = defaultdict(lambda: np.zeros(generator.shape))
    for index, row in tqdm(data.iterrows()):
        # if index == 10:
        #   break
        text = data[column][index]
        embeddings = generator.generate(text)

    # TODO benzer olan ilacın embedding değerini vererek dene
        # embedding check none type
        if embeddings is None or len(embeddings) == 0:
            sum_of_embeddings = np.zeros(generator.shape)
        else:
            sum_of_embeddings = np.sum(embeddings, axis=0)
        # column_embeddings_dict[row['id']] = sum_of_embeddings.reshape(1, -1) # 2d
        column_embeddings_dict[row['id']] = sum_of_embeddings
        # data.iloc[index][column+'_embedding']=sum_of_embeddings

    data[column+'_embedding'] = pd.Series(column_embeddings_dict.values())
    if (drop_column):
        data.drop([column], axis=1, inplace=True)
    # data[column+'_embedding'] = [column_embeddings_dict[row['name']] for index, row in data.iterrows()]
    return column_embeddings_dict



import pathlib
import numpy as np

import pandas as pd
from ..idf_helper import IDF

from ddi_fw.utils.zip_helper import ZipHelper
from .. import BaseDataset
from ..db_utils import create_connection

HERE = pathlib.Path(__file__).resolve().parent

'''
uses drugbank_db.zip
'''


class CustomDataset(BaseDataset):
    def __init__(self,  index_path, chemical_property_columns=['enzyme',
                                             'target',
                                             'pathway',
                                             'smile'],
                 drugbank_ids=[],
                 embedding_columns=['indication'],
                #  ner_columns = ['tui_description','cui_description', 'entities_description'],
                 ner_columns = ['tui','cui', 'entities'],
                 threshold_method='idf',
                 threshold_val=0
                 ):
        super().__init__(chemical_property_columns, embedding_columns, ner_columns, threshold_method, threshold_val)

        # self.embedding_columns = embedding_columns
        # self.columns = columns

        self.drugbank_ids = drugbank_ids

        zip_helper = ZipHelper()
        # zip_helper.extract(input_path=str(HERE.joinpath('drugbank_db.zip')), output_path=str(HERE))
        zip_helper.extract(input_path=str(HERE), output_path=str(HERE))
        # kwargs = {'index_path': str(HERE.joinpath('indexes'))}

        # db = HERE.joinpath('event.db')
        db = HERE.joinpath('drugbank.db')
        conn = create_connection(db)
        self.drugs_df = self.__select_all_drugs_as_dataframe__(conn)
        # dataframe'de olan ilaçlar arasındaki etkileşimleri bulmak için
        # self.drugbank_ids = self.drugs_df['id'].to_list()  ???
        self.ddis_df = self.__select_all_events__(conn)

        # self.index_path = kwargs.get('index_path')
        self.index_path = index_path

    def __select_all_events__(self, conn):
        param = tuple(self.drugbank_ids)
        cur = conn.cursor()
        query = f'''
            select
                _Interactions."index",
                drug_1_id,
                drug_1,
                drug_2_id,
                drug_2,
                mechanism_action, 
                interaction, 
                masked_interaction
            from _Interactions '''
        if len(self.drugbank_ids) > 0:
            query = query + \
                (f''' where drug_1_id in {format(param)} and drug_2_id in {format(param)}''')

        cur.execute(query)

        rows = cur.fetchall()

        headers = ['index', 'id1', 'name1', 'id2', 'name2',
                   'event_category', 'interaction', 'masked_interaction']
        df = pd.DataFrame(columns=headers, data=rows)
        return df

    # TODO tuis_description, entities_description, belli bir eşik değeri altında olanı ignore etmek lazım
    def __select_all_drugs_as_dataframe__(self, conn):
        param = tuple(self.drugbank_ids)
        query = f'''
            select 
            _Drugs."index",
            drugbank_id,
            _Drugs.name,
            description,
            synthesis_reference,
            indication,
            pharmacodynamics,
            mechanism_of_action,
            toxicity,
            metabolism,
            absorption,
            half_life,
            protein_binding,
            route_of_elimination,
            volume_of_distribution,
            clearance,
            smiles,
            smiles_morgan_fingerprint,
            enzymes_polypeptides,
            targets_polypeptides,
            pathways,
            tuis_description,
            cuis_description,
            entities_description
                    
            from _Drugs '''

        if len(self.drugbank_ids) > 0:
            query = query + f'''where 
                drugbank_id in {format(param)} and
                targets_polypeptides is not null and 
                enzymes_polypeptides is not null and 
                pathways is not null and 
                smiles_morgan_fingerprint is not null'''
        cur = conn.cursor()
        cur.execute(query)

        # pathway is absent

        rows = cur.fetchall()
        headers = ['index', 'id', 'name', 'description', 'synthesis_reference', 'indication', 'pharmacodynamics', 'mechanism_of_action', 'toxicity', 'metabolism', 'absorption', 'half_life',
                   'protein_binding', 'route_of_elimination', 'volume_of_distribution', 'clearance', 'smiles_notation', 'smile', 'enzyme', 'target', 'pathway',
                   'tui_description', 'cui_description', 'entities_description']
        df = pd.DataFrame(columns=headers, data=rows)
        df['smile'] = df['smile'].apply(lambda x:
                                        np.fromstring(
                                            x.replace(
                                                '\n', '')
                                            .replace('[', '')
                                            .replace(']', '')
                                            .replace('  ', ' '), sep=','))
        df['enzyme'] = df['enzyme'].apply(
            lambda x: x.split('|') if x is not None else [])
        df['target'] = df['target'].apply(
            lambda x: x.split('|') if x is not None else [])
        df['pathway'] = df['pathway'].apply(
            lambda x: x.split('|') if x is not None else [])
        df['tui_description'] = df['tui_description'].apply(
            lambda x: x.split('|') if x is not None else [])
        df['cui_description'] = df['cui_description'].apply(
            lambda x: x.split('|') if x is not None else [])
        df['entities_description'] = df['entities_description'].apply(
            lambda x: x.split('|') if x is not None else [])
        return df
